from Illuminate.Http.Request import Request
from Illuminate.Routing.RouteCollection import RouteCollection


class UrlGenerator:
    def __init__(
        self, routes: RouteCollection, request: Request, asset_root: str
    ) -> None:
        self._routes = routes
        self._asset_root = asset_root
        self._request = None

        self.set_request(request)

    def set_request(self, request: Request):
        self._request = request



from sys import argv
from typing import Self


class ArgvInput:
    def __init__(self):
        self.__inputs = argv[1:] if len(argv) > 1 else []

    def set_inputs(self, inputs: list = []):
        self.__inputs = inputs

    def get_all(self):
        return self.__inputs

    def get_options(self):
        return [input for input in self.__inputs if input.startswith("-")]

    def get_arguments(self):
        return [input for input in self.__inputs if not input.startswith("-")]

    def get_first_argument(self):
        arguments = self.get_arguments()

        return arguments[0] if arguments else None

    @classmethod
    def from_inputs(cls, inputs) -> Self:
        instance = cls()

        instance.set_inputs(inputs)

        return instance



from abc import abstractmethod
from typing import Any, Dict, Self
from Illuminate.Contracts.Support.JsonSerializable import JsonSerializable


class Collection(JsonSerializable):
    @abstractmethod
    def map(self, callback: Any) -> Self:
        raise NotImplementedError("Not Implemented")

    @abstractmethod
    def filter(self, callback: Any) -> Self:
        raise NotImplementedError("Not Implemented")

    @abstractmethod
    def all(self) -> Dict[Any, Any]:
        raise NotImplementedError("Not Implemented")

    @abstractmethod
    def count(self) -> int:
        raise NotImplementedError("Not Implemented")



from Illuminate.Validation.Rule import Rule


class Numeric(Rule):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def validate(self) -> bool:
        self.require_params_count(0)

        value = self.params.get("value")

        status, _ = self.is_numeric_value(value)

        return status

    def get_message(self) -> str:
        return "{field} must be a numeric value."



import inspect
import json

from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Helpers.Util import Util


class Logger:
    def __init__(self, app: Application) -> None:
        self.__app = app

    def log(self, info):
        print("logging", info)

    def dd(self, info):
        if isinstance(info, str):
            data = info
        elif isinstance(info, (int, float, bool)):
            data = str(info)
        elif isinstance(info, (list, dict)):
            data = json.dumps(Util.convert_values_to_string(info))
        elif callable(info):
            data = json.dumps(Util.convert_values_to_string(vars(info)))
        else:
            data = (
                json.dumps(Util.convert_values_to_string(vars(info)))
                if hasattr(info, "__dict__")
                else str(info)
            )

        return json.loads(data)

    @classmethod
    def caller_info(cls, stack_number=0):
        try:
            stack = inspect.stack()

            caller_frame = stack[stack_number]

            function_name = caller_frame.function

            module_name = caller_frame.frame.f_globals["__name__"]

            instance_method = caller_frame.frame.f_locals.get("self", None)

            if instance_method:
                return f"{module_name}.{instance_method.__class__.__name__}.{function_name}"
            else:
                return f"{module_name}.{function_name}"
        except Exception as e:
            print("logger exception", e)



from Illuminate.Auth.Access.Gate import Gate
from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Support.ServiceProvider import ServiceProvider


class AuthServiceProvider(ServiceProvider):
    def __init__(self, app: Application) -> None:
        self.__app = app

    def register(self):
        self.register_access_gate()

    def boot(self):
        pass

    def register_access_gate(self):
        self.__app.singleton("gate", lambda app: Gate(self.__app))



from Illuminate.Validation.Rule import Rule


class Integer(Rule):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def validate(self) -> bool:
        self.require_params_count(0)

        value = self.params.get("value")

        return isinstance(value, int)

    def get_message(self) -> str:
        return "{field} must be a integer value."



from pathlib import Path
from typing import Any, Self, Type
from Illuminate.Auth.AuthServiceProvider import AuthServiceProvider
from Illuminate.Foundation.Console.Input.ArgvInput import ArgvInput
from Illuminate.Foundation.Console.Output.ConsoleOutput import ConsoleOutput
from Illuminate.Events.Dispatcher import Dispatcher

from Illuminate.Events.EventServiceProvider import EventServiceProvider
from Illuminate.Foundation.Configuration.ApplicationBuilder import ApplicationBuilder
from Illuminate.Foundation.Http.Events.RequestReceived import RequestReceived
from Illuminate.Foundation.Providers.CommanderServiceProvider import (
    CommanderServiceProvider,
)
from Illuminate.Http.Request import Request
from Illuminate.Routing.ResponseFactory import ResponseFactory
from Illuminate.Log.LogServiceProvider import LogServiceProvider
from Illuminate.Routing.RoutingServiceProvider import RoutingServiceProvider

from Illuminate.Contracts.Http.Kernel import Kernel as HttpKernelContract
from Illuminate.Contracts.Console.Kernel import Kernel as ConsoleKernelContract
from Illuminate.Container.Container import Container
from Illuminate.Contracts.Container.Container import Container as ContainerContract
from Illuminate.Contracts.Foundation.Application import (
    Application as ApplicationContract,
)
from Illuminate.Contracts.Support.ServiceProvider import (
    ServiceProvider as ServiceProviderContract,
)

from Illuminate.Routing.Router import Router
from Illuminate.Contracts.Routing.Router import Router as RouterContract

from Illuminate.Support.Facades.Config import Config
from Illuminate.Validation.Factory import Factory as ValidationFactory
from Illuminate.Validation.ValidationServiceProvider import ValidationServiceProvider


class Application(Container, ApplicationContract):
    VERSION = "0.1.1"

    def __init__(self, base_path: str = None) -> None:
        super().__init__()

        self.__environment_path = None
        self.__environment_file = ".env"

        self.__base_path = None
        self.__app_path = None
        self.__config_path = None

        self.__database_path = None
        self.__public_path = None
        self.__resources_path = None
        self.__storage_path = None
        self.__lang_path = None
        self.__bootstrap_path = None

        self.__has_been_bootstrapped = False
        self.__running_in_console = False

        self.__booted = False
        self.__booting_callbacks = []
        self.__booted_callbacks = []

        self.__service_providers = {}

        self.__loaded_providers = {}

        self.__container_aliases = {
            "app": [Application, ApplicationContract, Container, ContainerContract],
            "request": [Request],
            "response": [ResponseFactory],
            "router": [Router, RouterContract],
            "events": [Dispatcher],
            "validator": [ValidationFactory],
        }

        if base_path:
            self.set_base_path(base_path)

        self._register_base_bindings()
        self._register_base_providers()
        self.register_container_aliases()

    @property
    def service_providers(self):
        return [provider for provider in self.__service_providers.values()]

    @property
    def loaded_providers(self):
        return self.__loaded_providers

    def __getitem__(self, key):
        return self.get_instance(key)

    def version(self):
        return self.VERSION

    def is_booted(self):
        return self.__booted

    def has_been_bootstrapped(self):
        return self.__has_been_bootstrapped

    def before_bootstraping(self, bootstrapper, callback):
        self.make("events").listen(f"bootstraping: {bootstrapper}", callback)

    def after_bootstraping(self, bootstrapper, callback):
        self.make("events").listen(f"bootstrapped: {bootstrapper}", callback)

    def bootstrap_with(self, bootstrappers):
        try:
            events = self.make("events")

            for bootstrapper in bootstrappers:
                events.dispatch(f"bootstraping: {bootstrapper}", [self])

                self.make(bootstrapper).bootstrap(self)

                events.dispatch(f"bootstrapped: {bootstrapper}", [self])

            self.__has_been_bootstrapped = True
        except Exception as e:
            raise e

    def set_base_path(self, base_path: str):
        self.__base_path = base_path

        self.__bind_path_in_container()

        return self

    def join_paths(self, base_path, path):
        return Path().joinpath(base_path, path)

    def base_path(self, path=""):
        return self.join_paths(self.__base_path, path)

    def app_path(self, path=""):
        return self.join_paths(
            self.__app_path if self.__app_path else self.base_path("app"),
            path,
        )

    def use_app_path(self, path: Path):
        self.__app_path = path
        self.instance("path", path)
        return self

    def config_path(self, path=""):
        return self.join_paths(
            self.__config_path if self.__config_path else self.base_path("config"),
            path,
        )

    def use_config_path(self, path: Path):
        self.__config_path = path
        self.instance("path.config", path)
        return self

    def database_path(self, path=""):
        return self.join_paths(
            (
                self.__database_path
                if self.__database_path
                else self.base_path("database")
            ),
            path,
        )

    def use_database_path(self, path: Path):
        self.__database_path = path
        self.instance("path.database", path)
        return self

    def public_path(self, path=""):
        return self.join_paths(
            self.__public_path if self.__public_path else self.base_path("public"),
            path,
        )

    def use_public_path(self, path: Path):
        self.__public_path = path
        self.instance("path.public", path)
        return self

    def resources_path(self, path=""):
        return self.join_paths(
            (
                self.__resources_path
                if self.__resources_path
                else self.base_path("resources")
            ),
            path,
        )

    def use_resources_path(self, path: Path):
        self.__resources_path = path
        self.instance("path.resources", path)
        return self

    def storage_path(self, path=""):
        return self.join_paths(
            self.__storage_path if self.__storage_path else self.base_path("storage"),
            path,
        )

    def use_storage_path(self, path: Path):
        self.__storage_path = path
        self.instance("path.storage", path)
        return self

    def lang_path(self, path=""):
        return self.join_paths(
            self.__lang_path if self.__lang_path else self.base_path("lang"),
            path,
        )

    def use_lang_path(self, path: Path):
        self.__lang_path = path
        self.instance("path.lang", path)
        return self

    def bootstrap_path(self, path=""):
        return self.join_paths(
            (
                self.__bootstrap_path
                if self.__bootstrap_path
                else self.base_path("bootstrap")
            ),
            path,
        )

    def use_bootstrap_path(self, path: Path):
        self.__bootstrap_path = path
        self.instance("path.bootstrap", path)
        return self

    def environment_path(self):
        return self.__environment_path if self.__environment_path else self.base_path()

    def use_environment_path(self, path: Path):
        self.__environment_path = path
        return self

    def environment_file(self):
        return self.__environment_file if self.__environment_file else ".env"

    def environment_file_path(self):
        return self.join_paths(
            self.environment_path(),
            self.environment_file(),
        )

    def load_environment_from(self, file):
        self.__environment_file = file
        return self

    def __bind_path_in_container(self):
        self.instance("path.base", self.base_path())
        self.instance("path.app", self.app_path())
        self.instance("path.config", self.config_path())
        self.instance("path.database", self.database_path())
        self.instance("path.public", self.public_path())
        self.instance("path.resources", self.resources_path())
        self.instance("path.storage", self.storage_path())
        self.instance("path.lang", self.lang_path())
        self.instance("path.bootstrap", self.bootstrap_path())

        self.use_bootstrap_path(self.bootstrap_path())
        self.use_lang_path(self.lang_path())

    def _register_base_bindings(self):
        self.instance("app", self)
        self.instance(Container, self)

    def _register_base_providers(self):
        self.register(AuthServiceProvider)
        self.register(EventServiceProvider)
        self.register(LogServiceProvider)
        self.register(RoutingServiceProvider)
        self.register(ValidationServiceProvider)
        self.register(CommanderServiceProvider)

    def register_configured_providers(self) -> Any:
        providers = Config.get("app.providers", [])

        for provider_class in providers:
            self.register(provider_class)

    def boot(self) -> Any:
        try:
            if self.is_booted():
                return

            self.fire_app_callbacks(self.__booting_callbacks)

            for service_provider in self.service_providers:
                self.boot_provider(service_provider)

            self.fire_app_callbacks(self.__booted_callbacks)

            self.__booted = True
        except Exception as e:
            raise e

    def fire_app_callbacks(self, callbacks):
        for callback in callbacks:
            callback(self)

    def register(self, provider_class: Type[ServiceProviderContract]):
        registered = self.get_provider(provider_class)

        if registered:
            return registered

        provider = provider_class(self)

        self.__service_providers[provider_class] = provider

        provider.register()

        self.__mark_as_registered(provider_class)

        if self.is_booted():
            self.boot_provider(provider)

        return provider

    def boot_provider(self, service_provider: ServiceProviderContract) -> Any:
        try:
            service_provider.call_booting_callbacks()
            service_provider.boot()
            service_provider.call_booted_callbacks()
        except Exception as e:
            raise e

    def booting(self, callback):
        self.__booting_callbacks.append(callback)

    def booted(self, callback):
        self.__booted_callbacks.append(callback)

        if self.is_booted():
            callback(self)

    def __mark_as_registered(self, base_key):
        self.__loaded_providers[base_key] = True

    def register_container_aliases(self):
        for key, aliases in self.__container_aliases.items():
            for alias in aliases:
                self.alias(key, alias)

    def get_provider(self, base_key):
        return self.__service_providers.get(base_key)

    def provider_is_loaded(self, base_key):
        return self.__loaded_providers.get(base_key)

    def detect_environment(self, callback):
        self.instance("env", callback())

    def set_running_in_console(self) -> Self:
        self.__running_in_console = True

        return self

    def running_in_console(self) -> bool:
        return self.__running_in_console

    def bound(self, abstract):
        return self.get_instance(abstract) is not None

    @classmethod
    def configure(cls, base_path: str = None) -> ApplicationBuilder:
        application = cls(base_path)

        return (
            ApplicationBuilder(application)
            .with_kernels()
            .with_events()
            .with_commands()
            .with_providers()
        )

    def handle_request(self, request: Request):
        self.make("events").dispatch(RequestReceived(request))

        kernel: HttpKernelContract = self.make(HttpKernelContract)

        response = kernel.handle(request)

        kernel.terminate(request, response)

        return response

    def handle_command(self, input: ArgvInput, silent=False):
        self.set_running_in_console()

        kernel: ConsoleKernelContract = self.make(ConsoleKernelContract)

        response = kernel.handle(input, ConsoleOutput(silent))

        kernel.terminate(input, response)

        return response



from typing import Any, Dict, List, Self
from Illuminate.Contracts.Http.Request import Request
from Illuminate.Routing.Controllers.HasMiddleware import Middleware


class Route:
    def __init__(
        self,
        methods,
        uri,
        action,
    ) -> None:
        self.methods = methods

        self.uri = uri

        self.action = action

        self.__computed_middleware = None

        self.__route_params: dict = {}

        self.__query_params: dict = {}

        self.__router = None

        self.__application = None

    @property
    def computed_middleware(self):
        return self.__computed_middleware

    @property
    def route_params(self):
        return self.__route_params

    @property
    def query_params(self):
        return self.__query_params

    @property
    def router(self):
        return self.__router

    @property
    def application(self):
        return self.__application

    def set_router(self, router) -> Self:
        self.__router = router

        return self

    def set_application(self, application) -> Self:
        self.__application = application

        return self

    def set_action(self, action: List[Any]) -> None:
        self.action = action

    def get_action(self) -> List[Any]:
        return self.action

    def bind(self, request: Request) -> Self:
        self.request = request

        return self

    def name(self, name: str):
        alias = self.action.get("as", "")

        if alias:
            self.action["as"] = f"{alias}{name}"
        else:
            self.action["as"] = name

        return self

    def run(self):
        try:
            controller = self.action.get("controller")

            if controller:
                action = self.__run_controller(controller)
            else:
                action = self.__run_callable()

            if not action:
                raise Exception("Invalid route action")

            dependencies = self.application.get_dependencies(action)

            return action(**dependencies)
        except Exception as e:
            raise e

    def __run_controller(self, controller):
        controller_object = self.application.make(controller)

        return getattr(controller_object, self.action["uses"])

    def __run_callable(self):
        return self.action["uses"]

    def gather_middleware(self):
        try:
            if not self.__computed_middleware:
                self.__computed_middleware = (
                    self.middleware() + self.controller_middleware()
                )

            return self.__computed_middleware
        except Exception as e:
            raise e

    def middleware(self, middleware=None):
        curent_middleware = self.action.get("middleware", [])

        if not middleware:
            return curent_middleware

        if isinstance(middleware, str):
            middleware = ",".split(middleware)

        if middleware:
            self.action["middleware"] = curent_middleware + middleware

        raise Exception("Invalid middleware")

    def controller_middleware(self):
        controller = self.action.get("controller")

        uses = self.action.get("uses")

        middleware_method = getattr(controller, "middleware", lambda: [])

        middleware = [
            (
                middleware
                if isinstance(middleware, Middleware)
                else Middleware(middleware)
            )
            for middleware in middleware_method()
        ]

        return [middleware.name for middleware in middleware if middleware.filter(uses)]

    def set_query_params(self, request: Request) -> Self:
        self.__query_params = request.query()

        return self

    def set_route_params(self, route_params: Dict[str, Any]) -> Self:
        self.__route_params = route_params

        return self



from typing import Dict
from Illuminate.Foundation.Console.Command import Command
from Illuminate.Foundation.Console.Input.ArgvInput import ArgvInput


class ContainerCommandLoader:
    def __init__(self, app, command_map: Dict[str, Command] = {}):
        self.app = app
        self.command_map: Dict[str, Command] = command_map

    def get_current_command(self, io: ArgvInput) -> Command:
        command_name = io.get_first_argument()

        if not command_name:
            return self.handle_invalid_command()

        try:
            return self.command_map[command_name]
        except KeyError:
            return self.handle_invalid_command()

    def handle_invalid_command(self) -> Command:
        return self.command_map["list"]



from abc import abstractmethod
from typing import Any


class Resolvable:
    @abstractmethod
    def resolve_for_display(self, resource, attribute) -> Any:
        raise NotImplementedError("Not Implemented")

    @abstractmethod
    def resolve(self, resource, attribute) -> Any:
        raise NotImplementedError("Not Implemented")



from pathlib import Path


def config(key=None):
    if not key:
        raise Exception("testing")

    parts = key.split(".")

    file_name = parts[0]

    file_path = Path(f"config/{file_name}.py")

    if not file_path.is_file():
        raise Exception("file not found")



import re

from abc import abstractmethod
from typing import TYPE_CHECKING, Any
from Illuminate.Foundation.Console.Input.ArgvInput import ArgvInput
from colorama import Fore, Back, Style

from Illuminate.Foundation.Console.Input.InputArgument import InputArgument
from Illuminate.Foundation.Console.Input.InputOption import InputOption
from Illuminate.Helpers.Util import Util

if TYPE_CHECKING:
    from Illuminate.Foundation.Console.Application import Application


class Command:
    name: str = ""
    signature: str = ""
    description: str = ""
    hidden = False
    silent = False

    def __init__(self):
        self.name = self.parse_command_name()

        self.segment = self._get_segment()

        self.options = self.merge_default_options(self.parse_command_options())

        self.arguments = sorted(
            self.parse_command_arguments(), key=lambda x: x["type"].value
        )

        self.input_options = []

        self.input_arguments = []

        self.commander = None

        self.application = None

    def get_default_options(self):
        return [
            [
                "help",
                "h",
                InputOption.VALUE_OPTIONAL,
                "Display help for the given command. When no command is given display help for the list command",
                None,
            ],
            [
                "quiet",
                "q",
                InputOption.VALUE_OPTIONAL,
                "Do not output any message",
                None,
            ],
            [
                "version",
                "v",
                InputOption.VALUE_OPTIONAL,
                "Display this application version",
                None,
            ],
        ]

    def merge_default_options(self, options: list):
        default_option = [
            self.transform_options(options) for options in self.get_default_options()
        ]

        options.extend(default_option)

        return options

    def _get_segment(self) -> str:
        items = self.name.split(":")

        if len(items) == 2:
            return items[0]

        return ""

    def set_silent(self, silent: bool = False):
        self.silent = silent

    @abstractmethod
    def handle(self):
        raise NotImplementedError("Not Implemented")

    def root_namespace(self):
        return self.application.app_path()

    def parse_command_name(self):
        return self.name if self.name else self.signature.split(" ")[0]

    def get_name_input(self):
        return self.argument("name").strip()

    def parse_command_arguments(self):
        if not self.signature:
            return [
                self.transform_arguments(arguments)
                for arguments in self.get_arguments()
            ]

        argument_pattern = r"\{(\w+)(\?)?(?:=(.*?))?\}"

        matches = re.findall(argument_pattern, self.signature)

        arguments = [
            {
                "name": match[0],
                "type": (
                    InputArgument.OPTIONAL if bool(match[1]) else InputArgument.REQUIRED
                ),
                "description": "",
                "value": match[2] if match[2] != "" else None,
            }
            for match in matches
        ]

        return arguments

    def parse_command_options(self):
        if not self.signature:
            return [self.transform_options(options) for options in self.get_options()]

        option_pattern = r"\{--(\w+)?\|?(\w+)?(?:=(.*?))?\}"

        matches = re.findall(option_pattern, self.signature)

        options = []

        for match in matches:
            short_name = match[0] if match[0] and match[1] else None
            long_name = match[0] if match[0] and not short_name else match[1]

            option_entry = {
                "short_name": short_name,
                "long_name": long_name,
                "name": self.get_option_name(long_name, short_name),
                "type": InputOption.VALUE_OPTIONAL,
                "description": "",
                "value": (
                    Util.to_boolean(match[2])
                    if isinstance(match[2], str) and len(match[2]) > 0
                    else False
                ),
            }

            options.append(option_entry)

        return options

    def set_commander(self, commander: "Application"):
        self.commander = commander
        self.application = commander.application

    def argument(self, name: str) -> Any:
        return self._get_input_value(name, self.arguments, "argument")

    def option(self, name: str):
        return self._get_input_value(name, self.options, "option")

    def _get_input_value(self, name: str, inputs: list, type: str):
        search = ["name"] if type == "argument" else ["short_name", "long_name"]

        for input in inputs:
            for key in search:
                if key in input and input[key] == name:
                    return input["value"]

        self.error(f"The '{name}' {type} does not exist.")
        exit()

    def get_option_name(self, long_name: str, short_name: str | None = None):
        short_name_title = f"-{short_name}, " if short_name else " "
        long_name_title = f"--{long_name}"

        return f"{short_name_title.rjust(4)}{long_name_title}"

    def parse_input(self, input: ArgvInput):
        arguments = input.get_arguments()

        self.input_options = input.get_options()

        self.input_arguments = arguments[1::]

        self.parse_options()

        self.parse_arguments()

    def validate(self):
        self.validate_arguments()
        self.validate_options()

    def validate_arguments(self):
        for argument in self.arguments:
            if argument["type"] == InputArgument.REQUIRED and not argument["value"]:
                self.error(f"{argument['name']} is required")

    def validate_options(self):
        for option in self.options:
            if option["type"] == InputOption.VALUE_REQUIRED and not option["value"]:
                self.error(f"{option['long_name']} is required")

    def parse_arguments(self):
        for argument in self.arguments:
            try:
                argument["value"] = self.input_arguments.pop(0)
            except IndexError:
                pass

    def parse_options(self):
        while True:
            try:
                input_option = self.input_options.pop(0)

                input_option_long_name = None
                input_option_short_name = None

                if input_option.startswith("--"):
                    input_option_data = input_option[2::].split("=")
                    input_option_long_name = input_option_data[0]
                elif input_option.startswith("-"):
                    input_option_data = input_option[1::].split("=")
                    input_option_short_name = input_option_data[0]
                else:
                    input_option_data = []

                assert len(input_option_data) <= 2, "Invalid option"

                input_option_value = (
                    True if len(input_option_data) == 1 else input_option_data[1]
                )

                for option in self.options:
                    if (
                        input_option_long_name
                        and option["long_name"] == input_option_long_name
                    ) or (
                        input_option_short_name
                        and option["short_name"] == input_option_short_name
                    ):
                        option["value"] = input_option_value
            except IndexError:
                break

    def get_arguments(self):
        return []

    def get_options(self):
        return []

    def line(self, message):
        print(f"{Fore.WHITE}{message}")

    def success(self, message):
        print(f"{Fore.GREEN}{message}")

    def info(self, message):
        print(f"{Fore.YELLOW}{message}")

    def error(self, message):
        horizontal_padding = 3

        vertical_padding = 1

        line_length = len(message) + (2 * horizontal_padding)

        horizontal_pad = " " * horizontal_padding

        padded_message = f"{horizontal_pad}{message}{horizontal_pad}"

        top_bottom_border = " " * line_length

        styled_text = f"{Back.RED}{Fore.WHITE}{Style.BRIGHT}{padded_message}"

        output = (
            f"\n" * vertical_padding
            + f"{Back.RED}{top_bottom_border}\n"
            + styled_text
            + "\n"
            + f"{Back.RED}{top_bottom_border}"
            + f"\n" * vertical_padding
        )

        exit(output)

    def new_line(self, total=1):
        for i in range(0, total):
            print("\n", end="")

    def transform_arguments(self, items: list):
        assert len(items) == 4, "Invalid argument payload"

        [name, type, description, value] = items

        if not name:
            raise Exception("name is required")

        if not isinstance(type, InputArgument):
            raise Exception("type must be enum type of InputArgument")

        return {
            "name": name,
            "type": type,
            "description": description,
            "value": value,
        }

    def transform_options(self, items: list):
        assert len(items) == 5, "Invalid argument payload"

        [long_name, short_name, type, description, value] = items

        if not long_name:
            raise Exception("long_name is required")

        if not isinstance(type, InputOption):
            raise Exception("type must be enum type of InputOption")

        return {
            "long_name": long_name,
            "short_name": short_name,
            "name": self.get_option_name(long_name, short_name),
            "type": type,
            "description": description,
            "value": value,
        }



from typing import Any, Dict, List, Union


class MiddlewareNameResolver:
    @classmethod
    def resolve(
        cls,
        name: str,
        map: Dict[str, Any],
        middleware_groups: Dict[str, List[Union[str, Any]]],
    ):
        if isinstance(name, str):
            if name in map:
                return map[name]

            if name in middleware_groups:
                resolved_group = []

                for item in middleware_groups[name]:
                    if isinstance(item, str):
                        resolved_middleware = cls.resolve(item, map, middleware_groups)

                        if resolved_middleware:
                            resolved_group.append(resolved_middleware)
                    else:
                        resolved_group.append(item)

                return resolved_group

        if hasattr(name, "handle"):
            return name

        return None



from abc import abstractmethod
from typing import Any, Type
from pathlib import Path

from Illuminate.Foundation.Console.Input.ArgvInput import ArgvInput
from Illuminate.Contracts.Container.Container import Container
from Illuminate.Http.RequestAdapter import RequestAdapter


class Application(Container):
    @property
    @abstractmethod
    def service_providers(self):
        """Returns the list of registered service providers."""
        pass

    @property
    @abstractmethod
    def loaded_providers(self):
        """Returns the loaded providers."""
        pass

    @abstractmethod
    def __getitem__(self, key: str) -> Any:
        """Gets the instance from the container."""
        pass

    @abstractmethod
    def is_booted(self) -> bool:
        """Checks if the application is booted."""
        pass

    @abstractmethod
    def has_been_bootstrapped(self) -> bool:
        """Checks if the application has been bootstrapped."""
        pass

    @abstractmethod
    def before_bootstraping(self, bootstrapper: str, callback: Any):
        """Registers a callback before bootstrapping."""
        pass

    @abstractmethod
    def after_bootstraping(self, bootstrapper: str, callback: Any):
        """Registers a callback after bootstrapping."""
        pass

    @abstractmethod
    def bootstrap_with(self, bootstrappers: list):
        """Bootstraps the application with the given bootstrappers."""
        pass

    @abstractmethod
    def base_path(self, path: str = "") -> Path:
        """Returns the base path of the application."""
        pass

    @abstractmethod
    def app_path(self, path: str = "") -> Path:
        """Returns the app path of the application."""
        pass

    @abstractmethod
    def use_app_path(self, path: Path):
        """Sets the application path."""
        pass

    @abstractmethod
    def config_path(self, path: str = "") -> Path:
        """Returns the config path of the application."""
        pass

    @abstractmethod
    def use_config_path(self, path: Path):
        """Sets the config path of the application."""
        pass

    @abstractmethod
    def bind(self, *args, **kwargs) -> None:
        """Binds an abstract to a concrete implementation."""
        pass

    @abstractmethod
    def singleton(self, *args, **kwargs) -> None:
        """Binds a singleton instance to the container."""
        pass

    @abstractmethod
    def make(self, *args, **kwargs) -> Any:
        """Resolves an instance from the container."""
        pass

    @abstractmethod
    def get_provider(self, base_key: str):
        """Gets the service provider by base key."""
        pass

    @abstractmethod
    def register_configured_providers(self) -> Any:
        """Registers providers configured in the application."""
        pass

    @abstractmethod
    def boot(self) -> Any:
        """Boots the application."""
        pass

    @abstractmethod
    def booting(self, callback: Any):
        """Registers a callback to be called during the booting process."""
        pass

    @abstractmethod
    def booted(self, callback: Any):
        """Registers a callback to be called after the boot process."""
        pass

    @abstractmethod
    def register(self, provider_class: Type):
        """Registers a service provider with the application."""
        pass

    @abstractmethod
    def boot_provider(self, service_provider: Any) -> Any:
        """Boots a service provider."""
        pass

    @abstractmethod
    def detect_environment(self, callback: Any):
        """Detects the application environment."""
        pass

    @abstractmethod
    def provider_is_loaded(self, base_key: str):
        """Gets the service provider by base key."""
        pass

    @abstractmethod
    def handle_request(self, request: RequestAdapter):
        """handle incoming request."""
        pass

    @abstractmethod
    def handle_command(self, input: ArgvInput):
        """handle command."""
        pass

    @abstractmethod
    def running_in_console(self) -> bool:
        """handle command."""
        pass

    @abstractmethod
    def set_running_in_console(self) -> "Application":
        """handle command."""
        pass



import math

from typing import Any, Callable, Dict, Generic, List, Optional, Self, Tuple, Union
from typing_extensions import TypeVar
from Illuminate.Collections.Arr import Arr
from Illuminate.Collections.Traits.EnumeratesValues import EnumeratesValues
from Illuminate.Helpers.Util import Util
from Illuminate.Contracts.Collections.Collection import Collection as CollectionContract

T = TypeVar("T")


class Collection(EnumeratesValues, Generic[T], CollectionContract):
    def __init__(self, items={}, *args, **kwargs) -> None:
        self._items = self._get_iterable_items(items)

    def all(self):
        return [(key, item) for key, item in self._items.items()]

    def count(self) -> int:
        return len(self._items)

    def is_empty(self) -> bool:
        return len(self._items) == 0

    def is_not_empty(self) -> bool:
        return not self.is_empty()

    def to_base(self) -> "Collection":
        return Collection(self._items)

    def values(self) -> Self:
        return self.__class__(self.to_list())

    def tap(self, callback: Callable[..., Any]) -> Self:
        callback(self)

        return self

    def prepend(self, value, key=None) -> Self:
        key = key if key else self._get_key()

        self._items = {key: value, **self._items}

        return self

    def filter(self, callback: Optional[Callable[..., Any]] = None) -> Self:
        if not callback:
            return self.__class__({key: value for key, value in self if value})

        self._check_is_callable(callback)

        results = {
            key: value
            for key, value in self
            if Util.callback_with_dynamic_args(callback, [value, key])
        }

        return self.__class__(results)

    def flat_map(self, callback: Callable[..., Any]) -> Self:
        return self.map(callback).flatten()

    def map(self, callback: Callable[..., Any]) -> Self:
        self._check_is_callable(callback)

        results = {
            key: Util.callback_with_dynamic_args(callback, [value, key])
            for key, value in self
        }

        return self.__class__(results)

    def map_with_keys(self, callback: Callable[..., Any]) -> Self:
        self._check_is_callable(callback)

        results = {}

        for key, value in self._items.items():
            assoc: dict = Util.callback_with_dynamic_args(callback, [value, key])

            for map_key, map_value in assoc.items():
                results[map_key] = map_value

        return self.__class__(results)

    def first(
        self, callback: Optional[Callable[[Any], bool]] = None, default=None
    ) -> Any:
        items = self._items

        if callback:
            items = self.filter(callback)._items

        return list(items.values())[0] if items else default

    def last(
        self, callback: Optional[Callable[[Any], bool]] = None, default=None
    ) -> Any:
        items = self._items

        if callback:
            items = self.filter(callback)._items

        return list(items.values())[-1] if items else default

    def push(self, value) -> Self:
        key = self._get_key()

        self._items[key] = value

        return self

    def pull(self, key, default=None) -> Self:
        return Arr.pull(self._items, key, default)

    def get(self, key, default=None) -> Self:
        return self._items[key] if key in self._items else default

    def unique(self, key: Optional[Callable] = None) -> Self:
        if not key:
            unique_values = set(self._items.values())

            return self.__class__(
                {key: value for key, value in enumerate(unique_values)}
            )

        callback = self._value_retriever(key)

        exists = []

        def is_unique(value, key):
            id = Util.callback_with_dynamic_args(callback, [value, key])

            if id in exists:
                return True

            exists.append(id)

        return self.reject(is_unique)

    def transform(self, callback: Callable[..., Any]) -> Self:
        self._items = self.map(callback)._items

        return self

    def concat(self, data: Union[List[Any], Dict[Any, Any], "Collection"] = []) -> Self:
        if isinstance(data, list):
            iterator = enumerate(data)
        elif isinstance(data, dict):
            iterator = data.items()
        elif isinstance(data, Collection):
            iterator = data.all()
        else:
            raise Exception(
                "Invalid items iterator, must be list, dict, or Collection object."
            )

        results = self.__class__(self._items)

        if not data:
            return results

        for key, item in iterator:
            results.push(item)

        return results

    def flatten(self, depth=math.inf) -> Self:
        flattened_items = Arr.flatten(self, depth)

        return self.__class__(flattened_items)

    def group_by(self, callback_or_string: Union[Callable[..., Any], str]):
        group_callback = self._value_retriever(callback_or_string)

        results: Dict[Any, Collection] = {}

        for key, value in self:
            group_keys = Util.callback_with_dynamic_args(group_callback, [value, key])

            def get_group_keys(group_keys):
                if isinstance(group_keys, str):
                    return [group_keys]

            group_keys = get_group_keys(group_keys) or [""]

            for group_key in group_keys:
                results.setdefault(group_key, self.__class__({})).push(value)

        return self.__class__(results)

    def sort_by(
        self,
        callback_or_string: Union[Callable[..., Any], str],
        descending: bool = False,
    ) -> Self:
        sort_callback = self._value_retriever(callback_or_string)

        sorted_items = self._sort_items(
            self.all(),
            key=lambda item: Util.callback_with_dynamic_args(
                sort_callback, [item[1], item[0]]
            ),
            descending=descending,
        )

        return self.__class__(sorted_items)

    def sort(
        self,
        sort_callback: Optional[Callable[..., Any]] = None,
        descending: bool = False,
    ) -> Self:
        if sort_callback:
            self._check_is_callable(sort_callback)

            return self.sort_by(sort_callback, descending)

        sorted_items = self._sort_items(
            self.all(), key=lambda item: str(item[1]), descending=descending
        )

        return self.__class__(sorted_items)

    def sort_keys(self, descending: bool = False) -> Self:
        sorted_items = self._sort_items(
            self.all(), key=lambda item: str(item[0]), descending=descending
        )

        return self.__class__(sorted_items)

    def _sort_items(
        self,
        items: List[Tuple[Any, Any]],
        key: Optional[Callable[[Tuple[Any, Any]], Any]] = None,
        descending: bool = False,
    ) -> Dict[Any, Any]:
        return dict(sorted(items, key=key, reverse=descending))

    def _get_iterable_items(self, data: Any) -> Dict[Any, Any]:
        try:
            if isinstance(data, Collection):
                return data._items
            elif (
                isinstance(data, list)
                and len(data)
                and all([isinstance(item, tuple) and len(item) == 2 for item in data])
            ):
                return dict(data)
            elif isinstance(data, (list, tuple)):
                return {key: item for key, item in enumerate(data)}
            elif isinstance(data, dict):
                return data
            else:
                return {0: data}
        except Exception as e:
            raise Exception("Invalid Items passed to collection")



from Illuminate.Contracts.Foundation.Application import Application


class Facade(type):
    app = None

    def __getattr__(cls, attribute, *args, **kwargs):
        facade_accessor = cls.get_facade_accessor()

        abstract = cls.app.get_alias(facade_accessor)

        binding = cls.app.make(abstract)

        return getattr(binding, attribute)

    @classmethod
    def set_facade_application(cls, app: Application):
        cls.app = app



from abc import ABC, abstractmethod
from typing import Any, Dict, Callable, Optional


class Container(ABC):
    @abstractmethod
    def bind(self, key: str, binding_resolver: Callable) -> None:
        """Bind a key to a binding resolver."""
        pass

    @abstractmethod
    def singleton(self, key: str, binding_resolver: Callable) -> None:
        """Bind a key to a binding resolver as a singleton."""
        pass

    @abstractmethod
    def make(self, key: str, make_args: Optional[Dict[str, Any]] = None) -> Any:
        """Resolve and create an instance of the given key."""
        pass

    @abstractmethod
    def bound(self, key: str) -> bool:
        """Check if a key is bound in the container."""
        pass

    @abstractmethod
    def alias(self, abstract_alias: str, alias: str) -> None:
        """Create an alias for an abstract key."""
        pass

    @abstractmethod
    def instance(self, key: str, instance: Any) -> Any:
        """Register an instance with a specific key."""
        pass

    @abstractmethod
    def forget_binding(self, key: str) -> None:
        """Remove a binding from the container."""
        pass

    @abstractmethod
    def forget_instance(self, key: str) -> None:
        """Remove an instance from the container."""
        pass

    @abstractmethod
    def reset(self) -> None:
        """Reset the container, clearing all bindings and instances."""
        pass

    @abstractmethod
    def has(self, key: str) -> bool:
        """Check if a key exists in the container."""
        pass

    @abstractmethod
    def get_aliases(self) -> Dict[str, str]:
        """Retrieve all aliases in the container."""
        pass

    @abstractmethod
    def get_bindings(self) -> Dict[str, Dict[str, Any]]:
        """Retrieve all bindings in the container."""
        pass

    @abstractmethod
    def get_instances(self) -> Dict[str, Any]:
        """Retrieve all instances in the container."""
        pass

    @abstractmethod
    def get_resolved(self) -> Dict[str, bool]:
        """Retrieve the resolved status of bindings."""
        pass



from abc import ABC
from typing import List, Self, Union


class Middleware:
    def __init__(
        self, name: str, only: List[str] = [], exclude: List[str] = []
    ) -> None:
        self.__name = name
        self.__only = only
        self.__exclude = exclude

    @property
    def name(self):
        return self.__name

    @property
    def only(self):
        return self.__only

    @property
    def exclude(self):
        return self.__exclude

    def set_only(self, only: Union[str | List[str]] = []) -> Self:
        self.__only = only if isinstance(only, list) else [only]

        return self

    def set_exclude(self, exclude: Union[str | List[str]] = []) -> Self:
        self.__exclude = exclude if isinstance(exclude, list) else [exclude]

        return self

    def filter(self, method) -> bool:
        if self.__only:
            return method in self.__only
        elif self.__exclude:
            return method not in self.__exclude

        return True


class HasMiddleware(ABC):
    @classmethod
    def middleware(cls) -> List[Union[str, Middleware]]:
        raise NotImplementedError("Must implement middleware method")



import re

from typing import List, Optional
from Illuminate.Exceptions.RouteNotFoundException import RouteNotFoundException
from Illuminate.Routing.Route import Route
from Illuminate.Contracts.Http.Request import Request


class RouteCollection:
    def __init__(self) -> None:
        self.routes = {}
        self.all_routes = {}

    def add(self, route: Route) -> None:
        for method in route.methods:
            method_wise_routes = self.routes.setdefault(method, {})
            method_wise_routes[route.uri] = route

        self.all_routes[route.uri] = route

    def match(self, request: Request):
        routes = [route for route in self.routes.get(request.get_method(), {}).values()]

        route = self._match_against_routes(request, routes)

        return self._handle_matched_route(request, route)

    def _match_against_routes(self, request: Request, routes: List[Route]):
        matched_routes = []

        for route in routes:
            matched = self.__run_match(request, route)

            if matched:
                matched_routes.append(matched)

        return matched_routes[0] if matched_routes else None

    def _handle_matched_route(self, request: Request, route: Optional[Route]):
        if route:
            return route.bind(request)

        raise RouteNotFoundException(
            f"The route {request.get_url()} could not be found."
        )

    def __run_match(self, request: Request, route: Route):
        request_path = request.get_url().strip("/")

        route_uri = route.uri.strip("/")

        params = re.findall(r":(\w+)", route.uri)

        if not params:
            return route if route_uri == request_path else None

        pattern = re.sub(r":(\w+)", r"([^/]+)", route_uri)

        pattern = f"^{pattern}$"

        match = re.match(pattern, request_path)

        if match:
            route_params = dict(zip(params, match.groups()))

            return route.set_route_params(route_params).set_query_params(request)
        else:
            return None



import abc

from typing import Any, Dict, List, Literal, TYPE_CHECKING, Tuple, Union

from Illuminate.Collections.helpers import collect
from Illuminate.Validation.helpers import pascal_to_snake

if TYPE_CHECKING:
    from Illuminate.Validation.RulesMapper import RulesMapper


class Rule(abc.ABC):
    def __init__(self, *args, **kwargs) -> None:
        self.__args = collect(args)
        self.__kwargs = kwargs

        self.__params: dict = {}
        self.__data: dict = {}
        self.__rule_type: Literal["string", "object"] | None = None
        self.__rules_mapper: "RulesMapper" | None = None

    @property
    def args(self):
        return self.__args

    @property
    def kwargs(self):
        return self.__kwargs

    @property
    def rule_type(self):
        return self.__rule_type

    @property
    def params(self):
        return self.__params

    @property
    def data(self):
        return self.__data

    @property
    def rules_mapper(self):
        return self.__rules_mapper

    @property
    def rule_name(self):
        return pascal_to_snake(self.__class__.__name__)

    @abc.abstractmethod
    def validate(self):
        raise NotImplementedError

    def get_message(self):
        return None

    def get_formatted_message(
        self,
        rule_item: str,
        messages: Dict[str, str],
        data: Dict[str, str],
    ):
        field = self.params.get("field")

        if f"{field}.{rule_item}" in messages:
            message = messages[f"{field}.{rule_item}"]
        elif field in messages:
            message = messages[field]
        else:
            message = self.get_message()

        return self.__format_template_message(message)

    def __format_template_message(self, template: str) -> str:
        for key, value in self.params.items():
            placeholder = f"{{{key}}}"

            template = template.replace(placeholder, str(value))

        return template

    def set_meta(
        self,
        rules_mapper: "RulesMapper",
        rule_type: Literal["string", "object"],
        data: Dict[str, Any],
    ):
        self.__rules_mapper = rules_mapper
        self.__rule_type = rule_type
        self.__data = data

        return self

    def require_params_count(self, count):
        if len(self.args) < count:
            raise Exception(
                f"Validation rule {self.rule_name} requires at least {count} parameters."
            )

    def set_params(self, params):
        self.__params.update(params)

        return self

    def is_validatable(self, field: str, rule_item: str) -> bool:
        present = self.check_present_or_implicit(field, rule_item)

        nullable = self.is_not_null_if_marked_as_nullable(field, rule_item)

        optional = self.passes_optional_check(field, rule_item)

        return present and nullable and optional

    def is_implicit_rule(self, field: str, rule_item: str):
        return rule_item in self.rules_mapper.implicit_rules

    def check_present_or_implicit(self, field: str, rule_item: str):
        field_exists = self.field_exists(field)

        return field_exists or self.is_implicit_rule(field, rule_item)

    def is_not_null_if_marked_as_nullable(self, field: str, rule_item: str):
        is_implicit = self.is_implicit_rule(field, rule_item)

        has_nullable = self.has_rule(field, ["nullable"])

        return is_implicit or not has_nullable

    def passes_optional_check(self, field: str, rule_item: str):
        has_sometimes = self.has_rule(field, ["sometimes"])

        if not has_sometimes:
            return True

        return self.field_exists(field)

    def field_exists(self, field: str):
        try:
            self.data[field]

            return True
        except KeyError:
            return False

    def value_exists(self, value: str):
        return self.is_countable(value) and len(value) > 0

    def has_rule(self, field: str, rules: List[str]) -> bool:
        field_rules = self.get_field_rules(field)

        return all(rule in field_rules for rule in rules)

    def get_field_rules(self, field: str) -> list:
        rules = self.rules_mapper.rules[field]

        return list(rules.keys())

    def get_size(self) -> int | float | None:
        field = self.params.get("field")

        value = self.params.get("value")

        has_numeric = self.has_rule(field, ["numeric"])

        is_numeric, numeric_value = self.is_numeric_value(value)

        self.is_numeric = is_numeric and has_numeric

        if self.is_numeric:
            return numeric_value
        elif isinstance(value, list):
            return len(value)
        else:
            return len(str(value))

    def is_numeric_value(self, value: Any) -> Tuple[bool, Union[int | float | None]]:
        if isinstance(value, (int, float)):
            return True, value

        if isinstance(value, str):
            try:
                return True, float(value)
            except ValueError:
                return False, None

        return False, None

    def is_countable(self, value) -> bool:
        return isinstance(value, (list, tuple, dict, set, range, str, bytes))



from Illuminate.Support.Str import Str


class InteractsWithContentTypes:
    def is_json(self):
        content_type = (
            self.header("Content-Type") if self.header("Content-Type") else ""
        )

        return Str.contains(content_type, ["/json", "+json"])



import inspect
import operator

from typing import Any, Callable, Optional, TypeVar
from Illuminate.Support.HigherOrderTapProxy import HigherOrderTapProxy


T = TypeVar("T")
R = TypeVar("R")


def safe_eval_compare(key, oper, value):
    compare_ops = {
        "==": operator.eq,
        "!=": operator.ne,
        ">": operator.gt,
        ">=": operator.ge,
        "<": operator.lt,
        "<=": operator.le,
    }

    if oper in compare_ops:
        return compare_ops[oper](key, value)
    else:
        raise ValueError("Invalid operator")


def tap(value: T, callback: Optional[Callable[[T], R]] = None):
    if not callback:
        return HigherOrderTapProxy(value)

    callback(value)

    return value


def transform(
    value: T,
    callback: Optional[Callable[[T], R]] = None,
    default: Optional[Callable[[T], R]] = None,
):
    if value:
        return callback(value)

    if callable(default):
        return default(value)

    return default


def with_(value: T, callback: Optional[Callable[[T], R]] = None) -> T:
    if not callback:
        return value

    return callback(value)


def is_class(obj: Any):
    return inspect.isclass(obj)


def is_class_instance(obj: Any):
    if is_class(obj):
        return False

    return isinstance(obj, obj.__class__)



class RouteGroup:
    @classmethod
    def merge(cls, new_attributes, old_attributes, prepend_existing_prefix=True):
        merged_group = {**old_attributes, **new_attributes}

        merged_group["prefix"] = cls._format_prefix(
            new_attributes, old_attributes, prepend_existing_prefix
        )

        return merged_group

    @classmethod
    def _format_prefix(
        cls, new_attributes, old_attributes, prepend_existing_prefix=True
    ):
        old_prefix: str = old_attributes.get("prefix", "")

        new_prefix: str = new_attributes.get("prefix", "")

        if prepend_existing_prefix:
            if new_prefix:
                return old_prefix.strip("/") + "/" + new_prefix.strip("/")

            return old_prefix

        return new_prefix.strip("/") + "/" + old_prefix.strip("/")



from datetime import datetime
from typing import Optional

from Illuminate.Contracts.Foundation.Application import (
    Application as ApplicationContract,
)
from Illuminate.Foundation.Http.Events.RequestHandled import RequestHandled
from Illuminate.Support.Facades.App import App
from Illuminate.Contracts.Http.Request import Request
from Illuminate.Contracts.Http.Response import Response
from Illuminate.Contracts.Routing.Router import Router as RouterContract
from Illuminate.Foundation.Bootstrap.BootProviders import BootProviders
from Illuminate.Foundation.Bootstrap.HandleExceptions import HandleExceptions
from Illuminate.Foundation.Bootstrap.LoadConfiguration import LoadConfiguration
from Illuminate.Foundation.Bootstrap.LoadEnvironmentVariables import (
    LoadEnvironmentVariables,
)
from Illuminate.Foundation.Bootstrap.RegisterFacades import RegisterFacades
from Illuminate.Foundation.Bootstrap.RegisterProviders import RegisterProviders
from Illuminate.Foundation.Http.Middleware.HandlePrecognitiveRequests import (
    HandlePrecognitiveRequests,
)

from Illuminate.Pipeline.Pipeline import Pipeline


class Kernel:
    __middleware: list = []
    __middleware_groups: dict = {}
    __middleware_aliases: dict = {}
    __route_middleware: dict = {}

    def __init__(self, app: ApplicationContract, router: RouterContract) -> None:
        self.__app = app
        self.__router = router

        self.__middleware = self.middleware
        self.__middleware_groups = self.middleware_groups
        self.__middleware_aliases = self.middleware_aliases

        self.__bootstrappers = [
            LoadEnvironmentVariables,
            LoadConfiguration,
            HandleExceptions,
            RegisterFacades,
            RegisterProviders,
            BootProviders,
        ]

        self.__middleware_priorities = [
            HandlePrecognitiveRequests,
        ]

        self.request_started_at: Optional[datetime] = None

        self.__sync_middleware_to_router()

    @property
    def app(self):
        return self.__app

    @property
    def router(self):
        return self.__router

    @property
    def bootstrappers(self):
        return self.__bootstrappers

    @property
    def middleware_priorities(self):
        return self.__middleware_priorities

    @property
    def middleware(self):
        return self.__middleware

    @property
    def middleware_groups(self):
        return self.__middleware_groups

    @property
    def middleware_aliases(self):
        return self.__middleware_aliases

    @property
    def route_middleware(self):
        return self.__route_middleware

    def __sync_middleware_to_router(self):
        self.router.middleware_priorities(self.middleware_priorities)

        for key, middleware in self.middleware_groups.items():
            self.router.middleware_group(key, middleware)

        merged_middleware = {
            **self.route_middleware,
            **self.middleware_aliases,
        }

        for key, middleware in merged_middleware.items():
            self.router.alias_middleware(key, middleware)

    def handle(self, request: Request) -> Response:
        try:
            self.request_started_at = datetime.now()

            response = self.send_through_router(request)

            App.make("events").dispatch(RequestHandled(request, response))

            return response
        except Exception as e:
            App.make("events").dispatch(RequestHandled(request, Response()))

            raise e

    def send_through_router(self, request: Request):
        try:
            self.__app.instance("request", request)

            self.__bootstrap()

            return (
                Pipeline(self.__app)
                .send(request)
                .through(self.middleware)
                .then(lambda request: self.__dispatch_to_router(request))
            )
        except Exception as e:
            raise e

    def __dispatch_to_router(self, request):
        self.__app.instance("request", request)

        data = self.router.dispatch(request)

        return data

    def __bootstrap(self):
        try:
            if not self.__app.has_been_bootstrapped():
                self.__app.bootstrap_with(self.bootstrappers)
        except Exception as e:
            raise e

    def terminate(self, request: Request, response: Response):
        if self.__app.bound("request"):
            self.__app.forget_binding("request")

    def push_middleware(self, middleware):
        if middleware not in self.__middleware:
            self.__middleware.append(middleware)

        return self



from Illuminate.Foundation.Console.Command import Command
from colorama import Fore


class HelpCommand(Command):
    signature = "help {name}"
    description = "help for given command"
    hidden = True

    def handle(self):
        command = self._get_current_command()

        padding = 4

        self.info(f"Description:")

        self.line(" " * padding + command.description)

        self.new_line()

        self.info(f"Usage:")

        options = f" [options] [--] " if len(command.options) > 1 else " "

        arguments = [f"<{argument['name']}>" for argument in command.arguments]

        self.line(" " * padding + f"{command.name}{options}{' '.join(arguments)}")

        self.new_line()

        max_length = self._get_max_length(
            command.options,
            [argument["name"] for argument in command.arguments if not command.hidden],
        )

        description_start = padding + max_length + padding

        if len(command.arguments):
            self.info(f"Arguments:")

            for argument in command.arguments:
                description_text = Fore.WHITE + argument["description"]

                self.success(
                    " " * padding
                    + argument["name"].ljust(description_start + padding)
                    + description_text
                )

            self.new_line()

        self.info(f"Options:")

        for option in command.options:
            description_text = (
                Fore.WHITE + option["description"] if "description" in option else ""
            )

            self.success(
                " " * padding
                + option["name"].ljust(description_start + padding)
                + description_text
            )

        self.new_line()

        exit()

    def _get_current_command(self):
        name = self.get_name_input()

        return self.commander.command_map[name]

    def _get_max_length(self, options: list, arguments: list):
        max_option_length = max(len(option["name"]) for option in options)

        max_argument_length = (
            max(len(argument) for argument in arguments) if len(arguments) else 0
        )

        return max([max_option_length, max_argument_length])



from abc import ABC, abstractmethod


class ServiceProvider(ABC):
    @property
    @abstractmethod
    def booting_callbacks(self):
        """Returns the registered booting callbacks."""
        pass

    @property
    @abstractmethod
    def booted_callbacks(self):
        """Returns the registered booted callbacks."""
        pass

    @abstractmethod
    def register(self):
        """Registers services in the container."""
        pass

    @abstractmethod
    def boot(self):
        """Boots the services."""
        pass

    @abstractmethod
    def booting(self, callback):
        """Registers a booting callback."""
        pass

    @abstractmethod
    def booted(self, callback):
        """Registers a booted callback."""
        pass

    @abstractmethod
    def call_booting_callbacks(self):
        """Calls the registered booting callbacks."""
        pass

    @abstractmethod
    def call_booted_callbacks(self):
        """Calls the registered booted callbacks."""
        pass



from abc import ABC, abstractmethod
from typing import Any, Callable, List


class Gate(ABC):
    @abstractmethod
    def check(self, ability: str, arguments: List[Any] = []) -> bool:
        pass

    @abstractmethod
    def define(self, ability: str, callback: Callable[[Any], Any]) -> None:
        pass



from pathlib import Path
from typing import Any, Dict, Self


class Repository:
    def __init__(
        self, items: Dict[str, Any] = {}, config_path: str | Path = ""
    ) -> None:
        self.__items: Dict[str, Any] = items

        self.__config_path: Path = (
            config_path if isinstance(config_path, Path) else Path(config_path)
        )

    def __setitem__(self, key, value):
        self.set(key, value)

    def __getitem__(self, key):
        return self.get(key)

    @property
    def path(self):
        return str(self.__config_path)

    def has(self, key):
        return self.get(key) is not None

    def get(self, key, default=None):
        config = self.__items
        parts = key.split(".")

        for part in parts:
            if isinstance(config, dict) and part in config:
                config = config[part]
            else:
                return default

        return config

    def get_all(self):
        return self.__items

    def set(self, key, value) -> Self:
        keys = key.split(".")
        config = self.__items

        for k in keys[:-1]:
            if k not in config or not isinstance(config[k], dict):
                config[k] = {}

            config = config[k]

        config[keys[-1]] = value

        return self

    def get_files(self):
        items = []

        if self.__config_path.exists():
            items = [
                file
                for file in self.__config_path.iterdir()
                if file.is_file() and file.suffix == ".py"
            ]

        return items



from typing import Any, Dict, List, Union
from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Validation.Rule import Rule
from Illuminate.Validation.RulesMapper import RulesMapper
from Illuminate.Validation.ValidationResponse import ValidationResponse


class Validator:
    def __init__(
        self,
        data: Dict[str, Any],
        rules: Dict[str, List[Union[str | Rule]]],
        messages: Dict[str, Any] | None = None,
        attributes: Dict[str, Any] | None = None,
    ):
        self.data = data

        self.rules = rules

        self.messages = messages or {}

        self.attributes = attributes or {}

        self.__validate_payload()

        self.validation_response = ValidationResponse(self.data)

        self.rules_mapper = RulesMapper()

        self.rules_mapper.set_rules(self.rules, self.data)

    def set_application(self, app: Application):
        self.__app = app

    def validate(self) -> ValidationResponse:
        try:
            for field, rules in self.rules_mapper.rules.items():
                for rule_item, rule_executor in rules.items():
                    validatable = rule_executor.is_validatable(field, rule_item)

                    if validatable:
                        should_process_next = self.__process_validation(
                            field,
                            rule_item,
                            rule_executor,
                            self.messages,
                            self.data,
                        )

                        if not should_process_next:
                            break

            return self.validation_response.execute()
        except Exception as e:
            raise Exception(str(e))

    def __process_validation(
        self,
        field: str,
        rule_item: str,
        rule_executor: Rule,
        messages: Dict[str, Any],
        data: Dict[str, Any],
    ):
        validated = rule_executor.validate()

        if not validated:
            formatted_message = rule_executor.get_formatted_message(
                rule_item, messages, data
            )

            self.validation_response.set_error(
                rule_executor.params.get("field"), formatted_message
            )

            if rule_item in self.rules_mapper.implicit_rules:
                return False

        return True

    def get_params(self, field_name: str, rule_name: str) -> Dict[str, Any]:
        params = {}

        for field, rules in self.rules_mapper.rules.items():
            for rule, executor in rules.items():
                if field == field_name and rule == rule_name:
                    params = executor.params

        return params

    def __validate_payload(self):
        assert isinstance(self.data, dict), "Data is not a dictionary"

        assert isinstance(self.messages, dict), "Messages is not a dictionary"

        assert isinstance(self.rules, dict), "Field rule items is not a dictionary"

        assert all(
            all(isinstance(item, (str, Rule)) for item in value)
            for value in self.rules.values()
        ), "Not all items in lists are strings or instances of Rule"



from typing import Any, Callable, Optional, Self, TypeVar, Generic
from Illuminate.Contracts.Collections.Collection import Collection as CollectionContract
from Illuminate.Helpers.Util import Util

T = TypeVar("T")


class Conditionable(Generic[T], CollectionContract):
    def when(
        self,
        value=None,
        callback: Optional[Callable[..., Any]] = None,
        default: Optional[Callable[..., Any]] = None,
    ) -> Self:
        value = value(self) if callable(value) else value

        if value:
            return (
                Util.callback_with_dynamic_args(callback, [self, value])
                if callable(callback)
                else self
            )
        else:
            return (
                Util.callback_with_dynamic_args(default, [self, value])
                if callable(default)
                else self
            )



from typing import Any, Dict, List
from Illuminate.Collections.helpers import collect
from Illuminate.Contracts.Events import Dispatcher
from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Helpers.Util import Util
from Illuminate.Routing.Events.PreparingResponse import PreparingResponse
from Illuminate.Routing.Events.ResponsePrepared import ResponsePrepared
from Illuminate.Routing.Events.Routing import Routing
from Illuminate.Http.Request import Request
from Illuminate.Pipeline.Pipeline import Pipeline
from Illuminate.Routing.MiddlewareNameResolver import MiddlewareNameResolver
from Illuminate.Routing.Route import Route
from Illuminate.Routing.RouteCollection import RouteCollection
from Illuminate.Routing.RouteGroup import RouteGroup
from Illuminate.Support.helpers import tap


class Router:
    def __init__(self, app: Application, events: Dispatcher) -> None:
        self.__app = app

        self.__events = events

        self._route_collection = RouteCollection()

        self.current_route = None

        self.current_request = None

        self.__middleware = {}

        self.__middleware_groups = {}

        self.__middleware_priorities = []

        self.__registered_paths = []

        self._group_stack = []

    @property
    def app(self):
        return self.__app

    @property
    def group_stack(self):
        return self._group_stack

    @group_stack.setter
    def group_stack(self, group_stack):
        self._group_stack = group_stack

    @property
    def events(self):
        return self.__events

    @property
    def route_collection(self):
        return self._route_collection

    def get_middleware(self):
        return self.__middleware

    def get_middleware_groups(self):
        return self.__middleware_groups

    def get_middleware_priorities(self):
        return self.__middleware_priorities

    def middleware_priorities(self, middleware_priorities: List[Any]):
        self.__middleware_priorities = middleware_priorities

        return self

    def middleware(self, middleware: List[Any]):
        self.__set_middleware(middleware)

        return self

    def middleware_group(self, key, middleware):
        self.__middleware_groups[key] = middleware

        return self

    def alias_middleware(self, key, middleware):
        self.__middleware[key] = middleware
        return self

    def get(self, uri, action):
        return self.add_route(["GET", "HEAD"], uri, action)

    def post(self, uri, action):
        return self.add_route(["POST"], uri, action)

    def put(self, uri, action):
        return self.add_route(["PUT"], uri, action)

    def patch(self, uri, action):
        return self.add_route(["PATCH"], uri, action)

    def delete(self, uri, action):
        return self.add_route(["DELETE"], uri, action)

    def option(self, uri, action):
        return self.add_route(["OPTION"], uri, action)

    def add_route(self, methods, uri, action):
        route = self._create_route(methods, uri, action)

        self._route_collection.add(route)

        return self

    def _create_route(self, methods, uri, action):
        action = self._convert_to_controller_action(action)

        route = self.new_route(
            methods,
            self._prefix(uri),
            action,
        )

        if self.has_group_stack():
            self._merge_group_attributes_into_routes(route)

        return route

    def _merge_group_attributes_into_routes(self, route: Route):
        route.set_action(
            self.merge_with_last_group(route.get_action(), False),
        )

    def _convert_to_controller_action(self, action):
        if Util.is_function(action):
            return {"uses": action}

        if isinstance(action, list) and len(action) == 2:
            controller_class, controller_action = action

            return {
                "controller": controller_class,
                "uses": controller_action,
            }

        if hasattr(action, "__call__"):
            return {
                "controller": action,
                "uses": "__call__",
            }

        raise Exception("BadMethodCallException")

    def _prefix(self, uri: str):
        last_prefix = self.get_last_group_prefix()

        trimmed_last_prefix = last_prefix.strip("/")

        trimmed_uri = uri.strip("/")

        return f"{trimmed_last_prefix}/{trimmed_uri}"

    def get_last_group_prefix(self):
        if self.has_group_stack():
            last_group_stack = self.group_stack[-1]

            return last_group_stack.get("prefix", "")

        return ""

    def has_group_stack(self):
        return bool(self.group_stack)

    def new_route(self, methods, uri, action):
        return Route(methods, uri, action).set_router(self).set_application(self.app)

    def dispatch(self, request: Request):
        self.current_request = request

        data = self.dispatch_to_route(request)

        return data

    def dispatch_to_route(self, request: Request):
        return self._run_route(request, self._find_route(request))

    def _run_route(self, request: Request, route: Route):
        request.set_route_resolver(lambda: route)

        return self.prepare_response(
            request,
            self._run_route_within_stack(route, request),
        )

    def _run_route_within_stack(self, route: Route, request: Request):
        middleware = self.gather_route_middleware(route)

        output = (
            Pipeline(self.__app)
            .send(request)
            .through(middleware)
            .then(lambda request: self.prepare_response(request, route.run()))
        )

        return output

    @classmethod
    def to_response(cls, request: Request, response: Any):
        return response

    def prepare_response(self, request: Request, response: Any):
        if not isinstance(request, Request):
            response = request

        self.events.dispatch(PreparingResponse(request, response))

        return tap(
            self.to_response(request, response),
            lambda response: self.events.dispatch(ResponsePrepared(request, response)),
        )

    def gather_route_middleware(self, route):
        route_middleware = route.gather_middleware()

        resolved_middleware = self.__resolve_middlware(route_middleware)

        sorted_middleware = self.__sort_middleware_by_priorities(resolved_middleware)

        return sorted_middleware

    def __resolve_middlware(self, middleware: List[Any]):
        return (
            collect(middleware)
            .map(
                lambda name: MiddlewareNameResolver.resolve(
                    name, self.__middleware, self.__middleware_groups
                )
            )
            .filter(lambda name: name is not None)
            .flatten()
            .all()
        )

    def __sort_middleware_by_priorities(self, middleware: Dict[Any, Any]):
        priorities = []

        non_priorities = []

        for key, item in middleware:
            if item in self.__middleware_priorities:
                priorities.append(item)
            else:
                non_priorities.append(item)

        return priorities + non_priorities

    def _find_route(self, request: Request):
        self.events.dispatch(Routing(request))

        route = self._route_collection.match(request)

        self.current_route = route

        route.set_application(self.__app)

        self.__app.instance(Route, route)

        return route

    def get_routes(self):
        return self._route_collection

    def register_path(self, path: str):
        self.__registered_paths.append(path)

    def get_registered_paths(self):
        return self.__registered_paths

    def group(self, attributes: Dict[str, Any], route_resolver: Any):
        group_routes = (
            route_resolver if isinstance(route_resolver, list) else [route_resolver]
        )

        for group_route in group_routes:
            self._update_group_stack(attributes)

            self._load_routes(group_route)

            self.group_stack.pop()

    def _update_group_stack(self, attributes):
        if self.has_group_stack():
            attributes = self.merge_with_last_group(attributes)

        self.group_stack.append(attributes)

    def merge_with_last_group(self, new_attributes, prepend_existing_prefix=True):
        last_group_stack = self.group_stack[-1]

        return RouteGroup.merge(
            new_attributes, last_group_stack, prepend_existing_prefix
        )

    def _load_routes(self, route_resolver):
        try:
            route_resolver(self)
        except Exception as e:
            raise e

    def __getattr__(cls, attribute, *args, **kwargs):
        if not cls.route_collection.all_routes:
            return

        last_route = list(cls.route_collection.all_routes.values())[-1]

        allowed_dynamic_methods = ["name", "middleware"]

        if attribute in allowed_dynamic_methods:
            return getattr(last_route, attribute)

        raise Exception("BadMethodCallException")



from Illuminate.Validation.Rule import Rule


class String(Rule):
    def __init__(self) -> None:
        super().__init__()

    def validate(self):
        self.require_params_count(0)

        value = self.params.get("value")

        return isinstance(value, str)

    def get_message(self):
        return "must be a string."



from typing import Callable
from Illuminate.Validation.Rule import Rule


class RequiredIf(Rule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def validate(self):
        status = (
            self.validate_object()
            if self.rule_type == "object"
            else self.validate_string()
        )

        if status:
            value = self.params.get("value")

            if self.value_exists(value):
                return self.value_exists(value)

        return True

    def validate_object(self) -> bool:
        self.require_params_count(1)

        self.set_params({"validator_object": self.args[0]})

        object_arg = self.args[0]

        if isinstance(object_arg, bool):
            return object_arg

        if isinstance(object_arg, Callable):
            return object_arg()

    def validate_string(self) -> bool:
        self.require_params_count(2)

        self.set_params(
            {"required_if_key": self.args[0], "required_if_value": self.args[1]}
        )

        required_if_key = self.params.get("required_if_key")

        required_if_value = self.params.get("required_if_value")

        expression = f"'{self.data.get(required_if_key)}' == {repr(required_if_value)}"

        return eval(expression)

    def get_message(self):
        if self.rule_type == "string":
            return "{field} is required when {required_if_key} is {required_if_value}"

        return "{field} is required"



from abc import ABC, abstractmethod
from datetime import datetime
from Illuminate.Contracts.Http.Request import Request
from Illuminate.Contracts.Http.Response import Response


class Kernel(ABC):
    @property
    @abstractmethod
    def middleware(self):
        """Returns the list of middlewares."""
        pass

    @property
    @abstractmethod
    def bootstrappers(self):
        """Returns the list of bootstrappers."""
        pass

    @property
    @abstractmethod
    def request_started_at(self) -> datetime:
        """Returns the request start time."""
        pass

    @abstractmethod
    def handle(self, request: Request):
        """Handles the incoming request."""
        pass

    @abstractmethod
    def terminate(self, request: Request, response: Response):
        """Handles the incoming request."""
        pass

    @abstractmethod
    def push_middleware(self):
        """Bootstraps the application if it hasn't been bootstrapped."""
        pass



import inspect

from typing import Any, Dict
from Illuminate.Support.helpers import is_class, is_class_instance


class Dispatcher:
    def __init__(self, app) -> None:
        self.__app = app
        self.__listeners: Dict[Any, Any] = {}

    def dispatch(self, event, args=[]):
        event_name = self.parse_event_name(event)

        callbacks = self.__listeners.get(event_name, [])

        for callback in callbacks:
            self.handle_callback(callback, event, args)

    def listen(self, event, callback):
        event_name = self.parse_event_name(event)

        self.__listeners.setdefault(event_name, [])

        self.__listeners[event_name].append(callback)

    def handle_callback(self, callback, event, args):
        if inspect.isclass(callback):
            callback(*args).handle(event)
        else:
            callback(*args)

    def parse_event_name(self, event):
        if isinstance(event, str):
            return event

        if is_class(event):
            return event.__name__

        if is_class_instance(event):
            return event.__class__.__name__

        raise Exception("Invalid event name", event)



import json

from Illuminate.Http.RequestAdapter import RequestAdapter
from Illuminate.Http.ServerBag.WSGIServer import WSGIServer


class WSGIRequestAdapter(RequestAdapter):
    def __init__(self, request: WSGIServer):
        self.request: WSGIServer = request

    def get_url(self):
        return self.request.path

    def get_full_url(self):
        return self.request.get_full_path()

    def query_data(self):
        return self.request.GET.dict()

    def post_data(self):
        try:
            return json.loads(self.request.body.decode("utf-8"))
        except json.JSONDecodeError:
            return {}



from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Routing.CallableDispatcher import CallableDispatcher
from Illuminate.Routing.Contracts.CallableDispatcher import (
    CallableDispatcher as CallableDispatcherContract,
)
from Illuminate.Routing.Contracts.ControllerDispatcher import (
    ControllerDispatcher as ControllerDispatcherContract,
)
from Illuminate.Routing.ControllerDispatcher import ControllerDispatcher
from Illuminate.Routing.Redirector import Redirector
from Illuminate.Routing.UrlGenerator import UrlGenerator

from Illuminate.Support.ServiceProvider import ServiceProvider
from Illuminate.Routing.Router import Router
from Illuminate.Http.Request import Request
from Illuminate.Routing.ResponseFactory import ResponseFactory


class RoutingServiceProvider(ServiceProvider):
    def __init__(self, app: Application) -> None:
        self.__app = app

    def register(self):
        self._register_router()
        self._register_url_generator()
        self._register_redirector()
        self._register_http_request()
        self._register_http_response()
        self._register_response_factory()
        self._register_callable_dispatcher()
        self._register_controller_dispatcher()

    def boot(self):
        pass

    def _register_router(self):
        def lambda_function(app: Application):
            return Router(app, app.make("events"))

        self.__app.singleton("router", lambda_function)

    def _register_url_generator(self):
        def lambda_function(app: Application):
            router = app.make("router")

            config = app.make("config")

            routes = router.get_routes()

            app.instance("routes", routes)

            return UrlGenerator(
                routes,
                self.__app.rebinding("request", self._request_rebinder()),
                config["app.asset_url"],
            )

        self.__app.singleton("url", lambda_function)

    def _request_rebinder(self):
        return lambda app, request: app.make("url").set_request(request)

    def _register_redirector(self):
        def lambda_function(app: Application):
            return Redirector(app, app.make("url"))

        self.__app.singleton("redirect", lambda_function)

    def _register_http_request(self):
        def lambda_function(app: Application):
            request = Request.capture(app)

            return request

        self.__app.singleton("request", lambda_function)

    def _register_http_response(self):
        def lambda_function(app: Application):
            return ResponseFactory(app)

        self.__app.singleton("response", lambda_function)

    def _register_response_factory(self):
        pass

    def _register_callable_dispatcher(self):
        def lambda_function(app: Application):
            return CallableDispatcher(app)

        self.__app.singleton(CallableDispatcherContract, lambda_function)

    def _register_controller_dispatcher(self):
        def lambda_function(app: Application):
            return ControllerDispatcher(app)

        self.__app.singleton(ControllerDispatcherContract, lambda_function)



class WSGIServer:
    __server = None

    def __init__(self, environ, start_response) -> None:
        self.environ = environ
        self.start_response = start_response

        self.scheme = environ["wsgi.url_scheme"]
        self.server_name = environ["SERVER_NAME"]
        self.server_port = environ["SERVER_PORT"]

        self.host = self.get_host()

        self.query_string = environ["QUERY_STRING"]
        self.method = environ["REQUEST_METHOD"]
        self.path = environ["PATH_INFO"]
        self.raw_path = environ["RAW_URI"]

        self.headers = {}
        self.cookies = {}

    @classmethod
    def create_server(cls, environ, start_response):
        server = WSGIServer(environ, start_response)
        WSGIServer.__server = server
        return server

    @classmethod
    def get_server(cls):
        return WSGIServer.__server

    def send(self):
        try:
            return ["test".encode("utf-8")]
        finally:
            print("done")

    def get_host(self):
        return (
            f"{self.scheme}://{self.server_name}"
            if self.server_port in [80, 443]
            else f"{self.scheme}://{self.server_name}:{self.server_port}"
        )



import math

from typing import Any, Dict, List, Union
from Illuminate.Collections.helpers import data_get
from Illuminate.Contracts.Collections.Collection import Collection as CollectionContract


class Arr:
    @classmethod
    def only(cls, items: Dict[Any, Any], keys: List[Any]):
        return {key: value for key, value in items.items() if key in keys}

    @classmethod
    def get(cls, items, key, default=None):
        return data_get(items, key, default)

    @classmethod
    def forget(cls, items, keys: List[str]):
        for key in keys:
            parts = key.split(".")
            current = items

            for part in parts[:-1]:
                if part in current:
                    current = current[part]
                else:
                    break
            else:
                if parts[-1] in current:
                    del current[parts[-1]]

    @classmethod
    def pull(cls, items, key, default=None):
        value = cls.get(items, key, default)

        cls.forget(items, key if isinstance(key, list) else [key])

        return value

    @classmethod
    def flatten(
        cls, items: Union[Dict[Any, Any], CollectionContract], depth: int = math.inf
    ) -> Dict[int, Any]:
        flattened = []

        def _flatten(current_items, current_depth):
            if isinstance(current_items, list):
                iterator = enumerate(current_items)
            elif isinstance(current_items, dict):
                iterator = current_items.items()
            elif isinstance(current_items, CollectionContract):
                iterator = current_items
            else:
                iterator = current_items

            for key, item in iterator:
                if isinstance(item, (CollectionContract, dict, list)) and (
                    depth == math.inf or current_depth < depth
                ):
                    _flatten(item, current_depth + 1)
                else:
                    flattened.append(item)

        _flatten(items, 0)

        return flattened



from Illuminate.Helpers.to_dict import to_dict


class HeaderBag:
    def __init__(self, headers) -> None:
        self.headers = {
            key.decode("utf-8"): value.decode("utf-8") for key, value in headers
        }

        self.cache_controller = {
            "cache-control": to_dict(self.headers.get("cache-control", "max-age=0"))
        }



import os

from pathlib import Path
from dotenv import load_dotenv
from Illuminate.Contracts.Foundation.Application import Application


class LoadEnvironmentVariables:
    def bootstrap(self, app: Application) -> None:
        environment_file_path = app.environment_file_path()
        self.__load_environment(environment_file_path, False)

        if os.getenv("APP_ENV"):
            env_path = f'{environment_file_path}.{os.getenv("APP_ENV")}'

            if Path(env_path).exists():
                app.load_environment_from(env_path)
                self.__load_environment(env_path, True)

    def __load_environment(self, env_path, override):
        load_dotenv(env_path, override=override)



import inspect

from functools import reduce


class Pipeline:
    def __init__(self, app) -> None:
        self.__app = app
        self.__passable = None
        self.__method = "handle"
        self.__pipes = []
        self.__output = None
        self.__current_index = 0

    def send(self, passable):
        self.__passable = passable

        return self

    def via(self, method):
        self.__method = method

        return self

    def pipe(self, pipe):
        self.__pipes.append(pipe)

        return self

    def through(self, pipes):
        self.__pipes = pipes

        reduce(self.__resolve_pipe, self.__pipes, True)

        return self

    def pipes(self):
        return self.__pipes

    def then(self, destination: callable):
        return destination(self.__output or self.__passable)

    def then_return(self):
        return self.then(lambda passable: passable)

    def __resolve_pipe(self, should_continue, current_pipe):
        caller = None

        if inspect.isclass(current_pipe):
            object = self.__app.make(current_pipe)
            caller = self.__get_caller(object)
        elif callable(current_pipe):
            caller = current_pipe
        else:
            caller = self.__get_caller(current_pipe)

        should_continue = should_continue and caller is not None

        return self.__manage_pipe(should_continue, caller)

    def __get_caller(self, obj):
        if hasattr(obj, self.__method):
            return getattr(obj, self.__method)
        else:
            return getattr(obj, "__call__")

    def __manage_pipe(self, should_continue, current_pipe):
        if should_continue:
            pipes = self.pipes()

            next_index = self.__current_index + 1

            if next_index == len(pipes):
                next_pipe = lambda name: self.__output
            else:
                next_pipe = lambda name: pipes[next_index]

            output = current_pipe(self.__passable, next_pipe)

            if callable(output):
                self.__current_index = next_index

                return True

            self.__output = output

        return False



from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Exceptions.Handler import Handler


class HandleExceptions:
    def bootstrap(self, app: Application) -> None:
        self.__app = app

        self.__app.singleton("exception_handler", lambda app: Handler(self.__app))



from abc import ABC, abstractmethod
from typing import Any


class RequestAdapter(ABC):
    request: Any

    @abstractmethod
    def get_url(self):
        raise NotImplementedError("Implement get_url method")

    @abstractmethod
    def get_full_url(self):
        raise NotImplementedError("Implement get_full_url method")

    @abstractmethod
    def get_method(self):
        raise NotImplementedError("Implement get_method method")

    @abstractmethod
    def get_user(self):
        raise NotImplementedError("Implement get_user method")

    @abstractmethod
    def query_data(self):
        raise NotImplementedError("Implement query_data method")

    @abstractmethod
    def post_data(self):
        raise NotImplementedError("Implement post_data method")

    @abstractmethod
    def form_data(self):
        raise NotImplementedError("Implement form_data method")

    @abstractmethod
    def json_data(self):
        raise NotImplementedError("Implement json_data method")

    @abstractmethod
    def files_data(self):
        raise NotImplementedError("Implement files_data method")

    @abstractmethod
    def headers_data(self):
        raise NotImplementedError("Implement headers_data method")

    @abstractmethod
    def sessions_data(self):
        raise NotImplementedError("Implement sessions_data method")

    @abstractmethod
    def cookies_data(self):
        raise NotImplementedError("Implement cookies_data method")



from Illuminate.Routing.UrlGenerator import UrlGenerator


class Redirector:
    def __init__(self, app, url_generator: UrlGenerator) -> None:
        self.__app = app
        self.__url_generator = url_generator

        self.url = ""

    def to(self, url):
        self.url = url

        return self



from typing import TYPE_CHECKING, Type

from Illuminate.Foundation.Support.Providers.RouteServiceProvider import (
    RouteServiceProvider as ServiceProvider,
)
from Illuminate.View.ViewFactory import ViewFactory

if TYPE_CHECKING:
    from Illuminate.Foundation.Application import Application


class ViewServiceProvider(ServiceProvider):
    def __init__(self, app: Type["Application"]) -> None:
        self.__app = app

    def register(self):
        self.__app.singleton("view", lambda app: ViewFactory(self.__app))

    def boot(self):
        pass



from Illuminate.Contracts.Foundation.Application import Application


class RegisterProviders:
    _merge: list = []

    def bootstrap(self, app: Application) -> None:
        self.__app = app

        self._merge_additional_providers(self.__app)

        self.__app.register_configured_providers()

    def _merge_additional_providers(self, app):
        config = app.make("config")

        providers = config.get("app.providers", [])

        for provider in providers:
            if provider not in self._merge:
                self._merge.append(provider)

        config.set("app.providers", self._merge)

        self._merge = []

    @classmethod
    def merge(cls, providers):
        for provider in providers:
            if providers not in cls._merge:
                cls._merge.append(provider)



from Illuminate.Foundation.Events.Dispatchable import Dispatchable
from Illuminate.Http.Request import Request


class Routing(Dispatchable):
    def __init__(self, request: Request):
        self.request = request



from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Contracts.Http.Kernel import Kernel as HttpKernelContract
from Illuminate.Foundation.Bootstrap.RegisterProviders import RegisterProviders
from Illuminate.Foundation.Http.Kernel import Kernel as HttpKernel
from Illuminate.Contracts.Console.Kernel import Kernel as ConsoleKernelContract
from Illuminate.Foundation.Console.Kernel import Kernel as ConsoleKernel


class ApplicationBuilder:
    def __init__(self, application: Application):
        self.__application = application

    def with_routing(self):
        return self

    def with_middleware(self):
        return self

    def with_exceptions(self):
        return self

    def with_kernels(self):
        self.__application.singleton(HttpKernelContract, HttpKernel)
        self.__application.singleton(ConsoleKernelContract, ConsoleKernel)

        return self

    def with_events(self):
        return self

    def with_commands(self):
        return self

    def with_providers(self, providers=[]):
        RegisterProviders.merge(providers)

        return self

    def create(self):
        return self.__application



import sys
import importlib

from abc import ABC, abstractmethod
from Illuminate.Foundation.Console.Application import Application as Commander


class ServiceProvider(ABC):
    booting_callbacks = {}
    booted_callbacks = {}

    @abstractmethod
    def register(self):
        pass

    @abstractmethod
    def boot(self):
        pass

    def booting(self, callback):
        self.booting_callbacks[self] = callback

    def booted(self, callback):
        self.booted_callbacks[self] = callback

    def call_booting_callbacks(self):
        callbacks = self.booting_callbacks.get(self)

        if callbacks:
            callbacks()

    def call_booted_callbacks(self):
        callbacks = self.booted_callbacks.get(self)

        if callbacks:
            callbacks()

    def load_routes_from(self, loader: str):
        try:
            if loader in sys.modules:
                importlib.reload(sys.modules[loader])
            else:
                importlib.import_module(loader)
        except Exception as e:
            raise e

    def commands(self, *commands):
        all_commands = []

        if len(commands) == 1 and isinstance(commands[0], list):
            all_commands.extend(commands[0])
        else:
            all_commands.extend(commands)

        Commander.starting(lambda commander: commander.resolve_commands(all_commands))



import importlib
import os

from typing import Any, Dict, List, Type
from Illuminate.Validation.Rule import Rule
from Illuminate.Validation.helpers import snake_to_pascal


class RulesMapper:
    def __init__(self) -> None:
        self.__rule_executor_types: Dict[str, Type[Rule]] = (
            self.__get_rule_executor_types()
        )

        self.__rules: dict[str, Dict[str, Rule]] = {}

        self.__implicit_rules = [
            "accepted",
            "accepted_if",
            "declined",
            "declined_if",
            "filled",
            "missing",
            "missing_if",
            "missing_unless",
            "missing_with",
            "missing_with_all",
            "present",
            "present_if",
            "present_unless",
            "present_with",
            "present_with_all",
            "required",
            "required_if",
            "required_if_accepted",
            "required_if_declined",
            "required_unless",
            "required_with",
            "required_with_all",
            "required_without",
            "required_without_all",
        ]

        self.__dependent_rules = [
            "after",
            "after_or_equal",
            "before",
            "before_or_equal",
            "confirmed",
            "different",
            "exclude_if",
            "exclude_unless",
            "exclude_with",
            "exclude_without",
            "gt",
            "gte",
            "lt",
            "lte",
            "accepted_if",
            "declined_if",
            "required_if",
            "required_if_accepted",
            "required_if_declined",
            "required_unless",
            "required_with",
            "required_with_all",
            "required_without",
            "required_without_all",
            "present_if",
            "present_unless",
            "present_with",
            "present_with_all",
            "prohibited",
            "prohibited_if",
            "prohibited_unless",
            "prohibits",
            "missing_if",
            "missing_unless",
            "missing_with",
            "missing_with_all",
            "same",
            "unique",
        ]

        self.__exclude_rules = [
            "exclude",
            "exclude_if",
            "exclude_unless",
            "exclude_with",
            "exclude_without",
        ]

        self.size_rules = [
            "between",
            "max",
            "min",
            "size",
            "gt",
            "lt",
            "gte",
            "lte",
        ]

        self.__numeric_rules = [
            "numeric",
            "integer",
            "decimal",
        ]

        self._initialized = True

    @property
    def rule_executor_types(self):
        return self.__rule_executor_types

    @property
    def rules(self):
        return self.__rules

    @property
    def implicit_rules(self):
        return self.__implicit_rules

    def __get_listable_rules(
        self, rules: Dict[str, str | List[Any]]
    ) -> Dict[str, List[Any]]:
        formatted_rules = {}

        for field, rule_item in rules.items():
            if isinstance(rule_item, str):
                formatted_rules[field] = rule_item.split("|")
            else:
                formatted_rules[field] = rule_item

        return formatted_rules

    def __get_rule_executor_types(self):
        rules_directory = os.path.join(os.path.dirname(__file__), "rules")

        rule_executor_types = {}

        for filename in os.listdir(rules_directory):
            if filename.endswith(".py") and filename != "__init__.py":
                rule_name = filename[:-3]

                class_name = snake_to_pascal(rule_name)

                try:
                    module = importlib.import_module(
                        f"Illuminate.Validation.rules.{rule_name}"
                    )

                    rule_class = getattr(module, class_name)

                    rule_executor_types[rule_name] = rule_class
                except (ImportError, AttributeError) as e:
                    print(f"Failed to load rule '{rule_name}': {e}")

        return rule_executor_types

    def set_rules(self, rules: Dict[str, List[Any]], data: Dict[str, Any]) -> None:
        listable_rules = self.__get_listable_rules(rules)

        for field, rule_items in listable_rules.items():
            value = data.get(field, None)

            params = {"field": field, "value": value}

            for rule_item in rule_items:
                rule_item_instance = None

                if isinstance(rule_item, Rule):
                    rule_item_instance = rule_item
                    rule_item_instance.set_meta(self, "object", data).set_params(params)

                if isinstance(rule_item, str):
                    rule_item_instance = self.__get_rule_executor(rule_item)
                    rule_item_instance.set_meta(self, "string", data).set_params(params)

                if not rule_item_instance:
                    raise Exception("Invalid Rule")

                self.__set_rule_executor(field, rule_item_instance)

    def __set_rule_executor(self, field: str, rule_item_instance: Rule) -> None:
        if field not in self.rules:
            self.rules[field] = {}

        rule_name = rule_item_instance.rule_name

        self.rules[field][rule_name] = rule_item_instance

    def __get_rule_executor(self, rule_item: str) -> Rule:
        try:
            rule_with_args = rule_item.split(":")

            rule = rule_with_args[0]

            rule_args = None if len(rule_with_args) == 1 else rule_with_args[1]

            rule_args = rule_args.split(",") if rule_args else []

            rule_executor_type = self.__rule_executor_types[rule]

            return rule_executor_type(*rule_args)
        except KeyError:
            raise Exception(f"Rule `{rule}` does not exists.")



import importlib

from Illuminate.Config.Repository import Repository
from Illuminate.Contracts.Foundation.Application import Application


class LoadConfiguration:
    def bootstrap(self, app: Application) -> None:
        self.__app = app

        config_path = self.__app.make("path.config")

        config = Repository({}, config_path)

        self.__app.instance("config", config)

        self.__load_configuration_files(config)

        self.__app.detect_environment(lambda: config.get("app.env", "production"))

    def __load_configuration_files(self, config):
        files = config.get_files()

        for file in files:
            file_name, file_content = self.__get_info(file)

            config.set(file_name, file_content)

    def __get_info(self, file):
        file_name = file.name[:-3]

        spec = importlib.util.spec_from_file_location(file_name, file)

        module = importlib.util.module_from_spec(spec)

        spec.loader.exec_module(module)

        return file_name, getattr(module, file_name)



import inspect

from abc import ABC
from typing import Any, Dict, Callable, Optional
from inspect import signature, getfullargspec
from Illuminate.Helpers.Util import Util
from Illuminate.Support.builtins import array_merge


class AttributeNotFound(Exception):
    pass


class BindingNotFound(Exception):
    pass


class BindingResolutionException(Exception):
    pass


class BuildingNewInstanceRequired(Exception):
    pass


class Container(ABC):
    def __init__(self) -> None:
        self.__bindings: Dict[str, Dict[str, Any]] = {}
        self.__instances: Dict[str, Any] = {}
        self.__resolved: Dict[str, bool] = {}
        self.__aliases: Dict[str, str] = {}
        self.__abstract_aliases: Dict[str, list] = {}
        self._global_before_resolving_callbacks: list = []
        self._before_resolving_callbacks: dict = {}
        self._global_resolving_callbacks: list = []
        self._resolving_callbacks: dict = {}
        self._global_after_resolving_callbacks: list = []
        self._after_resolving_callbacks: dict = {}

    def bind(self, key: str, binding_resolver: Callable) -> None:
        self.__bind(key, binding_resolver, False)

    def singleton(self, key: str, binding_resolver: Callable) -> None:
        self.__bind(key, binding_resolver, True)

    def make(self, key: str, parameters: Optional[Dict[str, Any]] = None) -> Any:
        return self.resolve(key, parameters)

    def resolve(self, abstract: str, parameters: Optional[Dict[str, Any]] = None):
        parameters = parameters or {}

        abstract = self.get_alias(abstract)

        self._fire_before_resolving_callbacks(abstract, parameters)

        concrete = self._get_contextual_concrete(abstract, parameters)

        needs_contextual_builds = parameters or concrete

        instance = self.get_instance(abstract)

        if instance and not needs_contextual_builds:
            return instance

        if not concrete:
            concrete = self._get_concrete(abstract, parameters)

        self._fire_resolving_callbacks(abstract, concrete)

        return concrete

    def _get_contextual_concrete(self, abstract, parameters):
        binding = self._find_in_contextual_bindings(abstract, parameters)

        if binding:
            return binding

        abstract_aliases = self.__abstract_aliases.get(abstract)

        if not abstract_aliases:
            return None

        for alias in abstract_aliases:
            binding = self._find_in_contextual_bindings(alias, parameters)

            if binding:
                return binding

    def _fire_before_resolving_callbacks(self, abstract, parameters):
        self._fire_before_callback_array(
            abstract, parameters, self._global_before_resolving_callbacks
        )

        for type, callbacks in self._before_resolving_callbacks.items():
            if type == abstract or (
                inspect.isclass(type)
                and inspect.isclass(abstract)
                and issubclass(type, abstract)
            ):
                self._fire_before_callback_array(abstract, parameters, callbacks)

    def _fire_before_callback_array(self, abstract, parameters, callbacks):
        for callback in callbacks:
            Util.callback_with_dynamic_args(callback, [abstract, parameters, self])

    def _fire_resolving_callbacks(self, abstract, concrete):
        self._fire_callback_array(concrete, self._global_resolving_callbacks)

        self._fire_callback_array(
            concrete,
            self._get_callbacks_for_type(abstract, concrete, self._resolving_callbacks),
        )

        self._fire_after_resolving_callbacks(abstract, concrete)

    def _fire_after_resolving_callbacks(self, abstract, concrete):
        self._fire_callback_array(concrete, self._global_after_resolving_callbacks)

        self._fire_callback_array(
            concrete,
            self._get_callbacks_for_type(
                abstract, concrete, self._after_resolving_callbacks
            ),
        )

    def _get_callbacks_for_type(self, abstract, concrete, callbacks_per_type):
        results = []

        for type, callbacks in callbacks_per_type.items():
            if type == abstract or (
                inspect.isclass(type) and isinstance(concrete, type)
            ):
                results = array_merge(results, callbacks)

        return results

    def _fire_callback_array(self, concrete, callbacks):
        for callback in callbacks:
            Util.callback_with_dynamic_args(callback, [concrete, self])

    def before_resolving(self, abstract, callback=None):
        if isinstance(abstract, str):
            abstract = self.get_alias(abstract)

        if callable(abstract) and not callback:
            self._global_before_resolving_callbacks.append(abstract)
        else:
            self._before_resolving_callbacks.setdefault(abstract, []).append(callback)

    def resolving(self, abstract, callback=None):
        if isinstance(abstract, str):
            abstract = self.get_alias(abstract)

        if callable(abstract) and not callback:
            self._global_resolving_callbacks.append(abstract)
        else:
            self._resolving_callbacks.setdefault(abstract, []).append(callback)

    def after_resolving(self, abstract, callback=None):
        if isinstance(abstract, str):
            abstract = self.get_alias(abstract)

        if callable(abstract) and not callback:
            self._global_after_resolving_callbacks.append(abstract)
        else:
            self._after_resolving_callbacks.setdefault(abstract, []).append(callback)

    def bind_if(self, key: str, binding_resolver: Callable) -> None:
        if not self.bound(key):
            self.__bind(key, binding_resolver, False)

    def bound(self, abstract: str) -> Optional[Any]:
        abstract = self.get_alias(abstract)

        binding = self.__bindings.get(abstract)
        instance = self.__instances.get(abstract)

        return instance if binding and instance else None

    def instance(self, abstract: str, instance: Any) -> Any:
        abstract = self.get_alias(abstract)
        return self.__make_instance(abstract, instance)

    def alias(self, abstract: str, alias: str) -> None:
        if abstract == alias:
            raise Exception(f"{abstract} cannot alias itself")

        self.__aliases[alias] = abstract
        self.__abstract_aliases.setdefault(abstract, []).append(alias)

    def get_alias(self, abstract: Any) -> Any:
        return (
            self.get_alias(self.__aliases.get(abstract))
            if abstract in self.__aliases
            else abstract
        )

    def get_aliases(self) -> Dict[str, str]:
        return self.__aliases

    def get_abstract_aliases(self) -> Dict[str, list]:
        return self.__abstract_aliases

    def get_bindings(self) -> Dict[str, Dict[str, Any]]:
        return self.__bindings

    def get_instance(self, key: str) -> Optional[Any]:
        return self.__instances.get(key)

    def get_instances(self) -> Dict[str, Any]:
        return self.__instances

    def get_resolved(self) -> Dict[str, bool]:
        return self.__resolved

    def __bind(self, abstract: str, binding_resolver: Callable, shared: bool) -> None:
        abstract = self.get_alias(abstract)

        self.__bindings[abstract] = {
            "abstract": abstract,
            "binding_resolver": binding_resolver,
            "shared": shared,
        }

    def __make_instance(self, abstract: str, instance: Any) -> Any:
        self.__instances[abstract] = instance
        return instance

    def _find_in_contextual_bindings(
        self, abstract: str, parameters: Dict[str, Any] = {}
    ) -> Optional[Any]:
        binding = self.__bindings.get(abstract)

        if not binding:
            return None

        binding_resolver = binding["binding_resolver"]
        shared = binding["shared"]

        if shared and (abstract in self.__instances):
            return self.__instances[abstract]

        instance = self.__resolve(abstract, binding_resolver, parameters)

        if shared:
            self.__instances[abstract] = instance

        return instance

    def _get_concrete(self, abstract: str, parameters: Dict[str, Any] = {}) -> Any:
        try:
            return self.__resolve(abstract, abstract, parameters)
        except Exception as e:
            raise BindingResolutionException(f"Error resolving class: {str(e)}")

    def __resolve(
        self,
        abstract: str,
        binding_resolver: Any,
        parameters: Dict[str, Any],
    ) -> Any:
        instance = None

        if inspect.isfunction(binding_resolver):
            total_params = len(signature(binding_resolver).parameters)

            if total_params >= 2:
                instance = binding_resolver(self, parameters)
            elif total_params == 1:
                instance = binding_resolver(self)
            else:
                instance = binding_resolver()

        if inspect.isclass(binding_resolver):
            dependencies = (
                parameters if parameters else self.get_dependencies(binding_resolver)
            )

            instance = binding_resolver(**dependencies)

        if instance:
            self.__resolved[abstract] = True
            return instance
        else:
            raise BindingResolutionException(
                f"Binding Resolution Exception for key {abstract} and class {binding_resolver}"
            )

    def get_dependencies(self, class_info) -> Dict[str, Any]:
        args_info = getfullargspec(class_info)

        return {
            arg: self.make(args_info.annotations[arg])
            for arg in args_info.args
            if arg != "self"
        }

    def get_resolved_dependencies(self, abstract: str) -> Dict[str, Any]:
        if abstract not in self.__resolved:
            raise BindingResolutionException(f"{abstract} is not resolved")

        return {abstract: self.__instances.get(abstract)}

    def clear_resolved(self) -> None:
        self.__resolved.clear()

    def has(self, abstract: str) -> bool:
        abstract = self.get_alias(abstract)

        return abstract in self.__bindings

    def forget_instance(self, abstract: str) -> None:
        abstract = self.get_alias(abstract)

        if abstract in self.__instances:
            del self.__instances[abstract]

    def forget_binding(self, abstract: str) -> None:
        abstract = self.get_alias(abstract)

        if abstract in self.__bindings:
            del self.__bindings[abstract]
            self.forget_instance(abstract)

    def reset(self) -> None:
        self.__bindings.clear()
        self.__instances.clear()
        self.__resolved.clear()
        self.__aliases.clear()
        self.__abstract_aliases.clear()



from typing import TYPE_CHECKING, Any, Type

from Illuminate.Contracts.Support.JsonSerializable import JsonSerializable
from Illuminate.Support.Facades.Event import Event

if TYPE_CHECKING:
    from Illuminate.Foundation.Application import Application


class ResponseFactory:
    def __init__(self, app: Type["Application"]) -> None:
        self.__app = app

        self.__content = ""
        self.__status = "200 OK"
        self.__headers = {"Content-type": "text/plain"}

    def get_status_code(self):
        return self.__status

    def get_headers(self):
        return [(key, value) for key, value in self.__headers.items()]

    def get_content(self):
        return self.__content

    def set_content(self, content: str):
        self.__content = content

        return self

    def set_status(self, status: str):
        self.__status = status

        return self

    def set_headers(self, key: str, value: Any):
        self.__headers[key] = value

        return self

    def send(self):
        Event.dispatch("response_sent", {"response": self})

        return self

    @classmethod
    def serialize(cls, data):
        if isinstance(data, JsonSerializable):
            return cls.serialize(data.json_serialize())
        elif (
            isinstance(data, list)
            and len(data)
            and all([isinstance(item, tuple) and len(item) == 2 for item in data])
        ):
            if all([isinstance(item[0], int) for item in data]):
                return cls.serialize(list(dict(data).values()))
            else:
                return cls.serialize(dict(data))
        elif isinstance(data, (list, tuple)):
            return [cls.serialize(item) for item in data]
        elif isinstance(data, (dict, set)):
            return {key: cls.serialize(value) for key, value in data.items()}
        else:
            return data



import re
import os

from pathlib import Path
from Illuminate.Foundation.Console.Command import Command
from Illuminate.Foundation.Console.Input.InputArgument import InputArgument


class GeneratorCommand(Command):
    def handle(self):
        try:
            self.generate_file_from_stub(
                self.get_stub(),
                self.get_path(),
                self.get_stub_vars(),
            )

            self.response()
        except Exception as e:
            self.error(str(e))

    def get_arguments(self):
        return [
            ["name", InputArgument.REQUIRED, f"The name of the {self.type}", None],
        ]

    def response(self):
        if not self.silent:
            self.success(f"{self.type} [{self.get_path()}] created successfully.")

    def get_path(self):
        namespace = self.get_default_namespace(self.root_namespace())

        namespace = str(namespace).rstrip("/")

        name = self.get_name_input()

        return f"{namespace}/{name}.py"

    def resolve_stub_path(self, stub: str, console_path="/"):
        return Path(console_path) / stub.strip("/")

    def get_stub(self):
        pass

    def get_stub_vars(self):
        return {}

    def get_default_namespace(self, root_namespace):
        return root_namespace

    def generate_file_from_stub(self, stub_file, main_file, variables={}):
        os.makedirs(os.path.dirname(main_file), exist_ok=True)

        if os.path.exists(main_file):
            if not self.silent:
                raise Exception(f"{self.type} already exists.")

        with open(stub_file, "r") as stub_file:
            stub_content = stub_file.read()

        modified_content = self.__replace_placeholders(stub_content, variables)

        with open(main_file, "w") as output_file:
            output_file.write(modified_content)

    def __replace_placeholders(self, content, replacements):
        def replacer(match):
            placeholder = match.group(1)

            return replacements.get(placeholder, match.group(0))

        pattern = r"\{\{(\w+)\}\}"

        return re.sub(pattern, replacer, content)



from Illuminate.Support.ServiceProvider import ServiceProvider
from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Validation.Factory import Factory


class ValidationServiceProvider(ServiceProvider):
    def __init__(self, app: Application) -> None:
        self.__app = app

    def register(self):
        self.__app.singleton("validator", lambda app: Factory(self.__app))

    def boot(self):
        pass



import re

from typing import Any
from Illuminate.Contracts.Container.Container import Container


class HandleCors:
    def __init__(self, container: Container) -> None:
        self.__container = container

    def handle(self, request, next) -> Any:
        config = self.__container.make("config")

        cors = config.get("cors", {})

        full_url = request.get_full_url()

        if not self.has_matching_path(cors, full_url):
            return next(request)

        return next(request)

    def has_matching_path(self, cors, full_url):
        paths = self.get_path_by_hosts(cors)

        for path in paths:
            return self.validate_path(full_url, path)

        return False

    def get_path_by_hosts(self, cors):
        return cors.get("paths")

    def validate_path(self, full_url, path):
        escaped_pattern = re.escape(path)
        regex_pattern = escaped_pattern.replace(r"\*", r".*")

        if re.search(regex_pattern, full_url):
            return True
        else:
            return False



import re

from Illuminate.Validation.Rule import Rule


class Email(Rule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.email_regex = r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$"

    def validate(self):
        self.require_params_count(0)

        value = self.params.get("value")

        if not value or not re.match(self.email_regex, str(value)):
            return False

        return True

    def get_message(self):
        return "The {field} must be a valid email address."



import time

from functools import lru_cache
from importlib import import_module
from typing import Any, Dict, List, Union


def use_module(ref: str):
    items = ref.split(".")

    class_name = items.pop()

    module_path = ".".join(items)

    module = import_module(module_path)

    return getattr(module, class_name)


def array_values(data: Dict[Any, Any]):
    if isinstance(data, list):
        return data
    elif isinstance(data, dict):
        return [item for key, item in data.items()]
    else:
        return data


from typing import Any, Dict, List, Union


def array_merge(
    main: Union[Dict[Any, Any], List[Any]],
    other: Union[Dict[Any, Any], List[Any]],
) -> Union[Dict[Any, Any], List[Any]]:
    if isinstance(main, list) and isinstance(other, list):
        return main + other
    elif isinstance(main, dict) and isinstance(other, dict):
        return {**main, **other}
    else:
        raise Exception(
            "Both arguments must be of the same type (either both lists or both dictionaries)"
        )


def array_merge_recursive(
    main: Union[Dict[Any, Any], List[Any]],
    other: Union[Dict[Any, Any], List[Any]],
) -> Union[Dict[Any, Any], List[Any]]:
    if isinstance(main, list) and isinstance(other, list):
        return main + other
    elif isinstance(main, dict) and isinstance(other, dict):
        merged = main.copy()

        for key, value in other.items():
            if (
                key in merged
                and isinstance(merged[key], dict)
                and isinstance(value, dict)
            ):
                merged[key] = array_merge_recursive(merged[key], value)
            elif (
                key in merged
                and isinstance(merged[key], list)
                and isinstance(value, list)
            ):
                merged[key] = merged[key] if not value else merged[key] + value
            elif key in merged and value is None:
                merged[key] = merged[key]
            else:
                merged[key] = value

        return merged
    else:
        raise Exception(
            "Both arguments must be of the same type (either both lists or both dictionaries)"
        )


def timed_lru_cache(seconds: int, maxsize: int = 128):
    """Decorator that caches results for a specified time (in seconds)."""

    def wrapper(func):
        func._cache_time = time.time()

        @lru_cache(maxsize=maxsize)
        def cached_func(*args, **kwargs):
            return func(*args, **kwargs)

        def wrapped_func(*args, **kwargs):
            if time.time() - func._cache_time > seconds:
                cached_func.cache_clear()
                func._cache_time = time.time()

            return cached_func(*args, **kwargs)

        return wrapped_func

    return wrapper



from Illuminate.Contracts.Http.Request import Request
from Illuminate.Contracts.Http.Response import Response


class RequestHandled:
    def __init__(self, request: Request, response: Response):
        self.request = request
        self.response = response



from Illuminate.Validation.Rule import Rule


class Max(Rule):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def validate(self):
        self.require_params_count(1)

        self.set_params({"max": self.args[0]})

        max = self.params.get("max")

        value_length = self.get_size()

        return value_length <= int(max)

    def get_message(self) -> str:
        message = "The {field} field must not be greater than {max}"

        return f"{message}." if self.is_numeric else f"{message} characters."



from Illuminate.Validation.Rule import Rule


class Required(Rule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def validate(self):
        self.require_params_count(0)

        value = self.params.get("value")

        return self.value_exists(value)

    def get_message(self):
        return "The {field} field is required"



from Illuminate.Validation.Rule import Rule


class Between(Rule):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def validate(self) -> bool:
        self.require_params_count(2)

        self.set_params({"small_value": self.args[0], "large_value": self.args[1]})

        small_value = self.params.get("small_value")

        large_value = self.params.get("large_value")

        value_length = self.get_size()

        return int(large_value) >= value_length >= int(small_value)

    def get_message(self) -> str:
        message = "The {field} field must be between {small_value} and {large_value}"

        return f"{message}." if self.is_numeric else f"{message} characters."



from Illuminate.Support.Facades.Facade import Facade
from Illuminate.Contracts.Foundation.Application import Application


class RegisterFacades:
    def bootstrap(self, app: Application) -> None:
        self.__app = app

        Facade.set_facade_application(self.__app)



class Serializable:
    def __init__(self, items=[]) -> None:
        self.items = items

    def __getitem__(self, index):
        return self.items[index]

    def __setitem__(self, index, value):
        self.items[index] = value

    def __delitem__(self, index):
        del self.items[index]

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.items})"



from jinja2 import Environment, FileSystemLoader


class ViewFactory:
    def __init__(self, app) -> None:
        self.__app = app
        self.__env = Environment(loader=FileSystemLoader("resources/views"))
        self.__content = ""

    def make(self, file, args={}):
        template = self.__env.get_template(f"{file}.html")
        self.__content = template.render(**args)
        return self

    def get_content(self):
        return self.__content



from typing import Any, Dict, Self
from Illuminate.Collections.helpers import collect
from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Http.Concerns.InteractsWithContentTypes import InteractsWithContentTypes
from Illuminate.Http.ServerBag.WSGIServer import WSGIServer
from Illuminate.Http.RequestAdapter import RequestAdapter
from Illuminate.Http.WSGIRequestAdapter import WSGIRequestAdapter
from Illuminate.Routing.Route import Route


class Request(InteractsWithContentTypes):
    def __init__(self, app: Application) -> None:
        self.app = app

        self.request_adapter: RequestAdapter = app.make("request_adapter")

        self.router = self.app.make("router")

    @classmethod
    def capture(cls, app):
        return cls.create_from_base(app)

    @classmethod
    def create_from_base(cls, app):
        server = WSGIServer.get_server()

        request_adapter = WSGIRequestAdapter(server)

        return cls.create_from(app, request_adapter)

    @classmethod
    def create_from(cls, app: Application, request_adapter: RequestAdapter):
        app.instance("request_adapter", request_adapter)

        return cls(app)

    def set_route_resolver(self, route_resolver) -> Self:
        self.route_resolver = route_resolver

        return self

    def current_route(self) -> Route:
        return self.router.current_route

    def get_route_params(self):
        current_route = self.current_route()

        return current_route.route_params

    def route_param(self, name, default=None):
        params = self.get_route_params()

        return params.get(name, default)

    def query_param(self, name, default=None):
        query = self.query()

        return query.get(name, default)

    def segment(self, number=1):
        if number < 1:
            raise None

        segments = self.segments()

        return segments.get(number - 1)

    def segments(self):
        url = self.get_url().strip("/")

        return collect(url.split("/") if isinstance(url, str) else [])

    def get_input_source(self) -> Dict[Any, Any]:
        if self.is_json():
            return self.json()

        method = self.get_method()

        if method in ["GET", "HEAD"]:
            return self.query()

        return self.form()

    def input(self) -> Dict[Any, Any]:
        get_input_source_data = self.get_input_source()

        query_data = self.query()

        data = {**get_input_source_data, **query_data}

        return data

    def all(self) -> Dict[Any, Any]:
        input_data = self.input()

        all_files_data = self.all_files()

        data = {**input_data, **all_files_data}

        return data

    def header(self, key) -> Any:
        return self.headers().get(key)

    def session(self, key) -> Any:
        return self.sessions().get(key)

    def cookie(self, key) -> Any:
        return self.cookies().get(key)

    def get_url(self) -> str:
        return self.request_adapter.get_url()

    def get_full_url(self) -> str:
        return self.request_adapter.get_full_url()

    def get_method(self) -> str:
        return self.request_adapter.get_method()

    def user(self) -> Any:
        return self.request_adapter.get_user()

    def json(self) -> Dict[Any, Any]:
        return self.request_adapter.json_data()

    def query(self) -> Dict[Any, Any]:
        return self.request_adapter.query_data()

    def post(self) -> Dict[Any, Any]:
        return self.request_adapter.post_data()

    def form(self) -> Dict[Any, Any]:
        return self.request_adapter.form_data()

    def all_files(self) -> Dict[Any, Any]:
        return self.request_adapter.files_data()

    def headers(self) -> Dict[Any, Any]:
        return self.request_adapter.headers_data()

    def sessions(self) -> Dict[Any, Any]:
        return self.request_adapter.sessions_data()

    def cookies(self) -> Dict[Any, Any]:
        return self.request_adapter.cookies_data()



from Illuminate.Collections.helpers import collect
from Illuminate.Contracts.Console.Kernel import Kernel as ConsoleKernelContract
from Illuminate.Foundation.Console.Command import Command
from colorama import Fore


class ListCommands(Command):
    name = "list"
    description = "List all commands"

    def handle(self):
        kernel = self.application.make(ConsoleKernelContract)

        commander = kernel.get_commander()

        padding = 4

        self.info(f"Flight Framework {self.commander.version}")

        self.new_line()

        self.info(f"Usage:")

        self.line(" " * padding + f"Commands [options] [arguments]")

        self.new_line()

        self.info(f"Options:")

        max_length = self._get_max_length(self.options, commander.command_map.items())

        description_start = padding + max_length + padding

        for option in self.options:
            description_text = Fore.WHITE + option["description"]

            self.success(
                " " * padding
                + option["name"].ljust(description_start + padding)
                + description_text
            )

        self.new_line()

        self.info(f"Available Commands:")

        commands = collect(commander.command_map).group_by(lambda item: item.segment)

        for group_key, group_commands in commands:
            if group_key:
                self.info(group_key.rjust(padding + 4))

            for key, item in group_commands:
                if not item.hidden:
                    description_text = Fore.WHITE + item.description

                    self.success(
                        " " * padding
                        + item.name.ljust(description_start + padding)
                        + description_text
                    )

        self.new_line()

        exit()

    def _get_max_length(self, options: list, commands: list):
        max_option_length = max(len(option["name"]) for option in options)

        max_command_length = max(
            len(item.name) for key, item in commands if not item.hidden
        )

        return max([max_option_length, max_command_length])



from datetime import datetime
from typing import Any
from Illuminate.Events.Dispatcher import Dispatcher
from Illuminate.Foundation.Console.Application import Application as Commander
from Illuminate.Foundation.Console.Input.ArgvInput import ArgvInput
from Illuminate.Foundation.Console.Output.ConsoleOutput import ConsoleOutput
from Illuminate.Contracts.Foundation.Application import (
    Application as ApplicationContract,
)
from Illuminate.Foundation.Bootstrap.BootProviders import BootProviders
from Illuminate.Foundation.Bootstrap.HandleExceptions import HandleExceptions
from Illuminate.Foundation.Bootstrap.LoadConfiguration import LoadConfiguration
from Illuminate.Foundation.Bootstrap.LoadEnvironmentVariables import (
    LoadEnvironmentVariables,
)
from Illuminate.Foundation.Bootstrap.RegisterFacades import RegisterFacades
from Illuminate.Foundation.Bootstrap.RegisterProviders import RegisterProviders


class Kernel:
    def __init__(self, app: ApplicationContract, events: Dispatcher) -> None:
        self.__app = app

        self.__events = events

        self.__bootstrappers = [
            LoadEnvironmentVariables,
            LoadConfiguration,
            HandleExceptions,
            RegisterFacades,
            RegisterProviders,
            BootProviders,
        ]

        self.commander: Commander | None = None

        self.commands: list = []

    @property
    def app(self):
        return self.__app

    @property
    def events(self):
        return self.__events

    @property
    def bootstrappers(self):
        return self.__bootstrappers

    def __bootstrap(self):
        try:
            if not self.__app.has_been_bootstrapped():
                self.__app.bootstrap_with(self.bootstrappers)
        except Exception as e:
            raise e

    def handle(self, input: ArgvInput, output: ConsoleOutput):
        try:
            self.command_started_at = datetime.now()

            self.__bootstrap()

            commander = self.get_commander()

            return commander.run(input, output)
        except Exception as e:
            print(e)

    def get_commander(self):
        if not self.commander:
            self.commander = (
                Commander(self.app, self.events, self.app.version())
                .resolve_commands(self.commands)
                .set_container_command_loader()
            )

        return self.commander

    def terminate(self, input: ArgvInput, response: Any):
        if self.commander:
            self.commander.terminate()



from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Validation.Validator import Validator


class Factory:
    def __init__(self, app: Application):
        self.__app = app

    def make(self, data, rules, messages=[], attributes=[]) -> Validator:
        validator = self._resolve(data, rules, messages, attributes)

        if self.__app:
            validator.set_application(self.__app)

        return validator

    def validate(self, data, rules, messages=[], attributes=[]):
        return self.make(data, rules, messages, attributes).validate()

    def _resolve(self, data, rules, messages=[], attributes=[]) -> Validator:
        return Validator(data, rules, messages, attributes)



from Illuminate.Http.CookieBag import CookieBag
from Illuminate.Http.HeaderBag import HeaderBag


class ASGIServer:
    scope = None
    receive = None
    send = None

    @classmethod
    def init(cls, scope, receive, send) -> None:
        cls.scope = scope
        cls.receive = receive
        cls.send = send

    def create(self):
        self.client_url = self.__parse_url(self.scope["client"])
        self.server_url = self.__parse_url(self.scope["server"])

        self.client_host = self.scope["client"][0]
        self.client_port = self.scope["client"][1]

        self.server_host = self.scope["server"][0]
        self.server_port = self.scope["server"][1]

        self.scheme = self.scope["scheme"]
        self.query_string = self.scope["query_string"]
        self.method = self.scope["method"]
        self.path = self.scope["path"]
        self.raw_path = self.scope["raw_path"]

        self.headers = HeaderBag(self.scope["headers"])
        self.cookies = CookieBag(self.headers.headers["cookie"])

    def __parse_url(self, items: tuple):
        return ":".join(str(item) for item in items)



import random
import re
import string
import unicodedata

from pluralizer import Pluralizer


class Str:
    _snake_cache = {}
    _kebab_cache = {}
    _camel_cache = {}
    _pascal_cache = {}
    _plural_cache = {}
    _singular_cache = {}
    _pluralizer = None

    @classmethod
    def random(cls, length=32) -> str:
        return "".join(random.choices(string.ascii_letters + string.digits, k=length))

    @classmethod
    def contains(cls, haystack, needles):
        if not haystack or not needles:
            return False

        return any(needle in haystack for needle in needles)

    @classmethod
    def get_pluralizer(cls):
        if not Str._pluralizer:
            Str._pluralizer = Pluralizer()

        return Str._pluralizer

    @classmethod
    def plural(cls, word):
        """Convert a word to its plural form using inflect library."""
        if word in Str._plural_cache:
            return Str._plural_cache[word]

        plural_form = Str.get_pluralizer().plural(word)
        Str._plural_cache[word] = plural_form
        return plural_form

    @classmethod
    def singular(cls, word):
        """Convert a word to its singular form using inflect library."""
        if word in Str._singular_cache:
            return Str._singular_cache[word]

        singular_form = Str.get_pluralizer().singular(word) or word
        Str._singular_cache[word] = singular_form
        return singular_form

    @classmethod
    def _normalize_and_clean(cls, string):
        """Normalize and clean a string for processing."""
        normalized_string = (
            unicodedata.normalize("NFKD", string)
            .encode("ascii", "ignore")
            .decode("utf-8")
        )
        cleaned_string = re.sub(r"[^a-z0-9\s]", "", normalized_string.lower())
        return cleaned_string

    @classmethod
    def snake(cls, value, delimiter="_"):
        """Convert a string to snake_case with a customizable delimiter."""
        if (cache := Str._snake_cache.get(value)) and delimiter in cache:
            return cache[delimiter]

        value = re.sub(
            r"(?<!^)(?=[A-Z])", " ", value
        )  # Add space before uppercase letters
        value = value.replace(" ", delimiter)
        result = value.lower()
        Str._snake_cache.setdefault(value, {})[delimiter] = result
        return result

    @classmethod
    def kebab(cls, value):
        """Convert a string to kebab-case."""
        if value in Str._kebab_cache:
            return Str._kebab_cache[value]

        result = Str.snake(value, "-")
        Str._kebab_cache[value] = result
        return result

    @classmethod
    def camel(cls, value):
        """Convert a string to camelCase."""
        if value in Str._camel_cache:
            return Str._camel_cache[value]

        words = re.findall(r"[A-Z][a-z]*|[a-z]+", value)
        result = (
            words[0].lower() + "".join(word.capitalize() for word in words[1:])
            if words
            else value
        )
        Str._camel_cache[value] = result
        return result

    @classmethod
    def pascal(cls, value):
        """Convert a string to PascalCase."""
        if value in Str._pascal_cache:
            return Str._pascal_cache[value]

        words = re.findall(r"[A-Z][a-z]*|[a-z]+", value)
        result = "".join(word.capitalize() for word in words)
        Str._pascal_cache[value] = result
        return result

    @classmethod
    def title(cls, value):
        """Convert a string to title case."""
        return value.title()

    @classmethod
    def lower(cls, value: str):
        """Convert a string to title case."""
        return value.lower()

    @classmethod
    def slug(cls, title, separator="-", language="en", dictionary={"@": "at"}):
        """Generate a URL-friendly slug from a given string."""
        title = Str._normalize_and_clean(title)
        flip = "_" if separator == "-" else "-"
        title = re.sub(rf"[{re.escape(flip)}]+", separator, title)

        for key, value in dictionary.items():
            title = title.replace(key, f"{separator}{value}{separator}")

        title = re.sub(r"[^a-z0-9\s]+", "", title.lower())
        title = re.sub(rf"[{re.escape(separator)}\s]+", separator, title)

        return title.strip(separator)



import inspect

from typing import Self
from Illuminate.Contracts.Foundation.Application import (
    Application as ApplicationContract,
)
from Illuminate.Events.Dispatcher import Dispatcher
from Illuminate.Foundation.Console.ContainerCommandLoader import ContainerCommandLoader
from Illuminate.Foundation.Console.Events.CommanderStarting import CommanderStarting
from Illuminate.Foundation.Console.Input.ArgvInput import ArgvInput
from Illuminate.Foundation.Console.Output.ConsoleOutput import ConsoleOutput
from Illuminate.Helpers.Util import Util
from Illuminate.Foundation.Console.Command import Command


class Application:
    bootstrappers: list = []
    command_map: dict = {}

    def __init__(self, app: ApplicationContract, events: Dispatcher, version):
        self.application = app

        self.events = events

        self.version = version

        self.events.dispatch(CommanderStarting(self))

        self.bootstrap()

        self.command_loader: ContainerCommandLoader | None = None

    @classmethod
    def starting(cls, callbacks):
        cls.bootstrappers.append(callbacks)

    def bootstrap(self):
        for bootstapper in self.bootstrappers:
            Util.callback_with_dynamic_args(bootstapper, [self])

    def run(self, input: ArgvInput, output: ConsoleOutput):
        try:
            assert self.command_loader is not None, "no command loader set"

            command = self.command_loader.get_current_command(input)

            command.set_silent(output.silent)

            command.parse_input(input)

            if command.option("help") or command.option("h"):
                return self.call("help", {"name": command.name})

            command.validate()

            action = getattr(command, "handle")

            return action()
        except Exception as e:
            print("error", e)

    def terminate(self):
        pass

    def call_silent(self, command: str, arguments: dict = {}):
        return self.call(command, arguments, True)

    def call(self, command: str, arguments: dict = {}, silent=False):
        data = [command, *list(arguments.values())]

        argv_input = ArgvInput.from_inputs(data)

        results = self.application.handle_command(argv_input, silent)

        return results

    def resolve_commands(self, commands) -> Self:
        all_commands = []

        if len(commands) == 1 and isinstance(commands[0], list):
            all_commands.extend(commands[0])
        else:
            all_commands.extend(commands)

        for command in all_commands:
            self.resolve(command)

        return self

    def resolve(self, command):
        if inspect.isclass(command) and issubclass(command, Command):
            return self.add(self.application.make(command))

        if isinstance(command, Command):
            return self.add(command)

        return self.add(self.application.make(command))

    def add(self, command: Command) -> Command:
        command.set_commander(self)

        self.command_map[command.name] = command

        return command

    def set_container_command_loader(self) -> Self:
        self.command_loader = ContainerCommandLoader(self.application, self.command_map)

        return self



from typing import TYPE_CHECKING, Type
from Illuminate.Events.Dispatcher import Dispatcher
from Illuminate.Support.ServiceProvider import ServiceProvider

if TYPE_CHECKING:
    from Illuminate.Foundation.Application import Application


class EventServiceProvider(ServiceProvider):
    def __init__(self, app: Type["Application"]) -> None:
        self.__app = app

    def register(self):
        self.__app.singleton("events", lambda app: Dispatcher(self.__app))

    def boot(self):
        pass



from Illuminate.Log.Logger import Logger


class LogManager:
    def __init__(self, app) -> None:
        self.__app = app
        self.__logger = Logger(self.__app)

    def log(self, info) -> None:
        return self.__logger.log(info)

    def dd(self, info) -> None:
        return self.__logger.dd(info)



import copy
import inspect
import types

from typing import Any, Callable, Dict, List


class Util:
    @classmethod
    def to_boolean(cls, value: Any):
        try:
            return bool(value)
        except:
            return value

    @classmethod
    def convert_values_to_string(cls, data):
        copied_data = copy.copy(data)

        def converter(obj):
            if isinstance(obj, dict):
                return {key: converter(value) for key, value in obj.items()}

            elif isinstance(obj, list):
                return [converter(value) for value in obj]

            elif callable(obj):
                return obj.__module__ + "." + obj.__name__

            return str(obj)

        return converter(copied_data)

    @classmethod
    def to_dict(cls, cookie):
        cookie_dict = {}

        for pair in cookie.split(";"):
            key, value = pair.strip().split("=")

            cookie_dict[key.strip()] = value.strip()

        return cookie_dict

    @classmethod
    def is_function(cls, item):
        return isinstance(item, types.FunctionType) or (
            isinstance(item, types.LambdaType) and item.__name__ == "<lambda>"
        )

    @staticmethod
    def get_callback_args_count(callback: Callable) -> int:
        sig = inspect.signature(callback)

        args = [
            p
            for p in sig.parameters.values()
            if p.kind in (p.POSITIONAL_OR_KEYWORD, p.POSITIONAL_ONLY)
        ]

        return len(args)

    @classmethod
    def callback_with_dynamic_args(
        cls, callback: Callable, args: List[Any] = [], kwargs: Dict[Any, Any] = {}
    ):
        try:
            args_count = cls.get_callback_args_count(callback)

            if not isinstance(args, list):
                raise Exception("Invalid args type, must be list of arguments")

            if args_count > len(args):
                raise Exception("Invalid arguments passed")

            build_args = args[:args_count]

            return callback(*build_args)
        except Exception as e:
            print(e)



from importlib import import_module
from typing import Callable
from Illuminate.Support.Facades.Route import Route
from Illuminate.Support.ServiceProvider import ServiceProvider


class RouteServiceProvider(ServiceProvider):
    def __init__(self) -> None:
        self.__load_routes_using: Callable[[], None]

    def routes(self, callback):
        self.__load_routes_using = callback

        return self

    def register(self):
        self.booted(self.load_routes)

    def load_routes(self):
        self.__load_routes_using()

        for registered_path in Route.get_registered_paths():
            import_module(registered_path)



from typing import TYPE_CHECKING, Any, List, TypeVar, Union

from Illuminate.Helpers.Util import Util
from Illuminate.Support.Facades.App import App

if TYPE_CHECKING:
    from Illuminate.Collections.Collection import Collection

T = TypeVar("T")


def event(*args, **kwargs):
    return App.make("events").dispatch(*args, **kwargs)


def collect(items: T = {}, *args, **kwargs) -> "Collection[T]":
    from Illuminate.Collections.Collection import Collection

    return Collection(items, *args, **kwargs)


def value(target, *args, **kwargs):
    return (
        Util.callback_with_dynamic_args(target, list(args), dict(kwargs))
        if callable(target)
        else target
    )


def data_get(target: Any, key: Union[List[str], str] = None, default=None):
    if not key or not isinstance(key, (list, str)):
        return target

    data_keys = [key] if isinstance(key, str) else key

    for data_key in data_keys:
        keys = data_key.split(".")
        current = target

        for item in keys:
            item = item.replace("->", ".")

            if isinstance(current, dict):
                current = current.get(item, default)
            else:
                current = getattr(current, item, default)

            if current is default:
                break

        if current is not default:
            return current

    return default



from typing import Any, Dict, List, Self


class ValidationResponse:
    def __init__(self, data: Dict[str, Any]) -> None:
        self.__initial_data = data
        self.__validated_data: dict = {}
        self.__validation_errors: dict = {}
        self.__validated = False

    @property
    def validated(self) -> bool:
        return self.__validated

    @property
    def data(self) -> Dict[str, Any]:
        return self.__validated_data

    @property
    def errors(self) -> Dict[str, List[str]]:
        return self.__validation_errors

    def set_error(self, field, value) -> None:
        field_errors = self.__validation_errors.setdefault(field, [])

        field_errors.append(value)

    def execute(self) -> Self:
        self.__validated = len(self.errors.keys()) == 0

        if self.__validated:
            self.__validated_data = self.__initial_data

        return self



import re


def snake_to_pascal(rule_name):
    """Converts snake_case to PascalCase."""
    return "".join(word.capitalize() for word in rule_name.split("_"))


def pascal_to_snake(class_name):
    """Converts PascalCase to snake_case."""
    snake_case = re.sub(r"(?<!^)(?=[A-Z])", "_", class_name).lower()

    return snake_case



from typing import Any, Callable, List, Self
from Illuminate.Collections.Enumerable import Enumerable
from Illuminate.Collections.helpers import data_get

from Illuminate.Conditionable.Conditionable import Conditionable
from Illuminate.Helpers.Util import Util
from Illuminate.Routing.ResponseFactory import ResponseFactory
from Illuminate.Support.helpers import safe_eval_compare


class EnumeratesValues(Conditionable, Enumerable):
    def __getitem__(self, index):
        return self._items[index]

    def __setitem__(self, index, value):
        self._items[index] = value

    def __delitem__(self, index):
        del self._items[index]

    def __iter__(self):
        return iter(self.all())

    def __len__(self):
        return len(self._items)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.json_serialize()})"

    def is_listable(self) -> bool:
        return all([isinstance(item, int) for item in self._items])

    def to_list(self) -> List[Any]:
        return list(self._items.values())

    def map_into(self, class_name) -> Self:
        return self.map(lambda value, key: class_name(value, key))

    def reject(self, callback=True) -> Self:
        return self.filter(
            lambda value, key: (
                not Util.callback_with_dynamic_args(callback, [value, key])
                if callable(callback)
                else self._bool_value(value) != self._bool_value(callback)
            )
        )

    def each(self, callback: Callable[[Any], Any]) -> Self:
        self._check_is_callable(callback)

        for key, value in self:
            status = Util.callback_with_dynamic_args(callback, [value, key])

            if status == False:
                break

        return self

    def every(self, data_key, data_operator=None, data_value=None) -> bool:
        if not data_operator and not data_value:
            self._check_is_callable(data_key)

            callback = self._value_retriever(data_key)

            for key, value in self:
                status = Util.callback_with_dynamic_args(callback, [value, key])

                if self._bool_value(status) == False:
                    return False

            return True

        return self.every(
            lambda value: safe_eval_compare(
                data_get(value, data_key), data_operator, data_value
            )
        )

    def where_instance_of(self, instance: Any) -> Self:
        return self.filter(lambda item: isinstance(item, instance))

    def partition(self, data_key, data_operator=None, data_value=None):
        passed = {}
        failed = {}

        if not data_operator and not data_value:
            self._check_is_callable(data_key)

            callback = self._value_retriever(data_key)

            for key, value in self:
                if Util.callback_with_dynamic_args(callback, [value, key]):
                    passed[key] = value
                else:
                    failed[key] = value

            return [self.__class__(passed), self.__class__(failed)]

        return self.partition(
            lambda value: safe_eval_compare(
                data_get(value, data_key), data_operator, data_value
            )
        )

    def json_serialize(self) -> Any:
        return ResponseFactory.serialize(
            self.to_list() if self.is_listable() else self._items
        )

    def _bool_value(self, value):
        return True if value else False

    def _value_retriever(self, value):
        if callable(value):
            return value

        return lambda item: data_get(item, value)

    def _check_is_callable(self, callback: Any) -> None:
        if not callable(callback):
            raise ValueError("Expected a callable")

    def _get_key(self):
        if not self.count():
            return 0

        keys = [key for key in self._items.keys() if isinstance(key, int)]

        if keys:
            return max(keys) + 1

        return 0



from collections.abc import Callable
from typing import Any
from Illuminate.Contracts.Foundation.Application import Application
from Illuminate.Contracts.Http.Request import Request


class HandlePrecognitiveRequests:
    def __init__(self, app: Application) -> None:
        self.__app = app

    def handle(self, request: Request, next: Callable[[Any], Any]):
        return next(request)



from Illuminate.Validation.Rule import Rule


class Accepted(Rule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.accepts = ["yes", "on", 1, "1", True, "true"]

    def validate(self):
        self.require_params_count(0)

        value = self.params.get("value")

        if self.value_exists(value):
            return value in self.accepts

        return False

    def get_message(self):
        return "The {field} field must be accepted"



from Illuminate.Validation.Rule import Rule


class Nullable(Rule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def validate(self):
        self.require_params_count(0)

        return True



from importlib import import_module
from pathlib import Path


def load_files(path_string: str):
    path = Path(path_string)

    files = [file for file in path.glob("*.py")]

    for file in files:
        module_path = f"{path}/{file.stem}"
        module_path = module_path.replace("/", ".")

        import_module(module_path, package=None)



from abc import abstractmethod
from typing import Any, Callable, Self
from Illuminate.Contracts.Collections.Collection import Collection as CollectionContract


class Enumerable(CollectionContract):
    @abstractmethod
    def map_into(self, class_name) -> Self:
        raise NotImplementedError("Not Implemented")

    @abstractmethod
    def reject(self, callback) -> Self:
        raise NotImplementedError("Not Implemented")

    @abstractmethod
    def each(self, callback: Callable[[Any], Any]) -> Self:
        raise NotImplementedError("Not Implemented")

    @abstractmethod
    def every(self, data_key, data_operator, data_value) -> int:
        raise NotImplementedError("Not Implemented")

    @abstractmethod
    def partition(self, data_key, data_operator, data_value) -> int:
        raise NotImplementedError("Not Implemented")



from typing import TYPE_CHECKING, Type

from Illuminate.Foundation.Console.HelpCommand import HelpCommand
from Illuminate.Foundation.Console.ListCommands import ListCommands
from Illuminate.Support.ServiceProvider import ServiceProvider

if TYPE_CHECKING:
    from Illuminate.Foundation.Application import Application


class CommanderServiceProvider(ServiceProvider):
    def __init__(self, app: Type["Application"]) -> None:
        self.__app = app

        self._commands = [
            ListCommands,
            HelpCommand,
        ]

    def register(self):
        self._register_commands(self._commands)

    def boot(self):
        pass

    def _register_commands(self, commands: list):
        self.commands(commands)



from Illuminate.Validation.Rule import Rule


class Sometimes(Rule):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def validate(self) -> bool:
        return True



from typing import Any, Callable, List
from Illuminate.Collections.helpers import collect
from Illuminate.Contracts.Auth.Access.Gate import Gate as GateContract
from Illuminate.Exceptions.UnauthorizedAccessException import (
    UnauthorizedAccessException,
)
from Illuminate.Helpers.Util import Util


class Gate(GateContract):
    def __init__(
        self,
        app,
        user_resolver=lambda: None,
        abilities={},
        policies={},
        before_callbacks: List[Callable[[Any], Any]] = [],
        after_callbacks: List[Callable[[Any], Any]] = [],
    ) -> None:
        self.__app = app
        self.user_resolver = user_resolver
        self.abilities = abilities
        self.policies = policies
        self.before_callbacks = before_callbacks
        self.after_callbacks = after_callbacks

    def before(self, callback: Callable[[Any], Any]):
        if callable(callback):
            self.before_callbacks.append(callback)

        return self

    def after(self, callback: Callable[[Any], Any]):
        if callable(callback):
            self.after_callbacks.append(callback)

        return self

    def define(self, ability: str, callback: Callable[[Any], Any]):
        if callable(callback):
            self.abilities[ability] = callback

    def check(self, abilities: list, arguments: List[Any] = []) -> bool:
        return collect(abilities).every(
            lambda ability: self.inspect(ability, arguments)
        )

    def authorize(self, ability: str, arguments: List[Any] = []):
        assert isinstance(ability, str)

        allowed = self.inspect(ability, arguments)

        if not allowed:
            raise UnauthorizedAccessException("Unauthorized")

    def inspect(self, ability: str | list, arguments: List[Any] = []) -> bool:
        results = self.raw(ability, arguments)

        return results

    def raw(self, ability: str, arguments: List[Any] = []) -> bool:
        arguments = arguments if isinstance(arguments, list) else [arguments]

        user = self.resolve_user()

        auth_callback = self.get_auth_callback(ability, arguments)

        if not auth_callback:
            return True

        return Util.callback_with_dynamic_args(auth_callback, [user, *arguments])

    def get_auth_callback(self, ability: str | list, arguments: List[Any] = []):
        auth_callback = None

        if len(arguments):
            policy_class = self.get_policy_for(arguments[0])

            if policy_class and hasattr(policy_class, ability):
                auth_callback = getattr(policy_class(), ability)

        if not auth_callback:
            ability_callback = self.abilities.get(ability)

            if ability_callback:
                auth_callback = ability_callback

        return auth_callback

    def has(self, abilities: List[str]) -> bool:
        return all(self.abilities.get(ability, None) for ability in abilities)

    def get_policy_for(self, instance_or_class):
        model_class = self.__get_model_class(instance_or_class)

        try:
            return self.policies[model_class]
        except KeyError:
            return None

    def policy(self, instance_or_class, policy):
        model_class = self.__get_model_class(instance_or_class)

        if model_class:
            self.policies[model_class] = policy

    def __get_model_class(self, instance_or_class):
        if isinstance(
            instance_or_class, (str, bool, int, float, list, dict, tuple, set)
        ):
            return None
        elif isinstance(instance_or_class, type):
            return instance_or_class
        else:
            return instance_or_class.__class__

    def for_user(self, user):
        return Gate(
            app=self.__app,
            user_resolver=lambda: user,
            abilities=self.abilities,
            policies=self.policies,
            before_callbacks=self.before_callbacks,
            after_callbacks=self.after_callbacks,
        )

    def resolve_user(self):
        return self.user_resolver()

    def resolve_policy(self, policy):
        return self.__app.make(policy)

    def allows(self, ability: str, arguments: List[Any]) -> bool:
        return self.check(ability, arguments)

    def denies(self, ability: str, arguments: List[Any]) -> bool:
        return not self.allows(ability, arguments)



from Illuminate.Validation.Rule import Rule


class AcceptedIf(Rule):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.accepts = ["yes", "on", 1, "1", True, "true"]

    def validate(self) -> bool:
        self.require_params_count(2)

        self.set_params(
            {"required_if_key": self.args[0], "required_if_value": self.args[1]}
        )

        required_if_key = self.params.get("required_if_key")

        required_if_value = self.params.get("required_if_value")

        expression = f"'{self.data.get(required_if_key)}' == {repr(required_if_value)}"

        status = eval(expression)

        if status:
            value = self.params.get("value")

            if self.value_exists(value):
                return value in self.accepts

            return False

        return True

    def get_message(self) -> str:
        return "{field} is required when {required_if_key} is {required_if_value}"
